"""Tests for the OAuth applications web API,."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.utils import six
from djblets.db.query import get_object_or_none
from djblets.testing.decorators import add_fixtures
from djblets.webapi.errors import DOES_NOT_EXIST
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.oauth.forms import ApplicationChangeForm
from reviewboard.oauth.models import Application
from reviewboard.site.models import LocalSite
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (oauth_app_item_mimetype,
                                                oauth_app_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_oauth_app_item_url,
                                           get_oauth_app_list_url)


def _compare_item(self, item_rsp, app):
    self.assertEqual(item_rsp['authorization_grant_type'],
                     app.authorization_grant_type)
    self.assertEqual(item_rsp['client_id'], app.client_id)
    self.assertEqual(item_rsp['client_secret'], app.client_secret)
    self.assertEqual(item_rsp['client_type'], app.client_type)
    self.assertEqual(item_rsp['id'], app.pk)
    self.assertEqual(item_rsp['name'], app.name)

    if app.redirect_uris:
        uris = {uri.strip() for uri in app.redirect_uris.split(',')}
    else:
        uris = set()

    self.assertEqual(set(item_rsp['redirect_uris']), uris)
    self.assertEqual(item_rsp['skip_authorization'], app.skip_authorization)
    self.assertEqual(item_rsp['links']['user']['title'], app.user.username)


@six.add_metaclass(BasicTestsMetaclass)
class ResourceListTests(ExtraDataListMixin, BaseWebAPITestCase):
    """Testing the OAuthApplicationResource list APIs."""

    resource = resources.oauth_app
    sample_api_url = 'oauth-apps/'

    fixtures = ['test_users']

    compare_item = _compare_item

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        if populate_items:
            if with_local_site:
                local_site = LocalSite.objects.get(name=local_site_name)
            else:
                local_site = None

            items = [
                Application.objects.create(user=user, local_site=local_site),
            ]
        else:
            items = []

        return (get_oauth_app_list_url(local_site_name=local_site_name),
                oauth_app_list_mimetype,
                items)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_filtered(self):
        """Testing the GET <URL> API only returns filtered applications"""
        admin = User.objects.get(username='admin')
        local_site = LocalSite.objects.get(pk=1)

        applications = set(filter(
            lambda a: a.local_site is None and a.user == self.user,
            self._make_applications([self.user, admin], local_site),
        ))

        rsp = self.api_get(get_oauth_app_list_url(),
                           {},
                           expected_mimetype=oauth_app_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(applications,
                         self._applications_from_response(rsp['oauth_apps']))

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_filtered_with_localsite(self):
        """Testing the GET <URL> API only returns filtered applications on a
        LocalSite
        """
        admin = User.objects.get(username='admin')
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)

        applications = self._make_applications(
            users=[self.user, admin],
            local_site=local_site,
            predicate=lambda a: (a.local_site == local_site and
                                 a.user == self.user),
        )

        rsp = self.api_get(get_oauth_app_list_url(local_site.name),
                           {},
                           expected_mimetype=oauth_app_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(applications,
                         self._applications_from_response(rsp['oauth_apps']))

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_superuser_get(self):
        """Testing the GET <URL> API as a superuser"""
        self.user = self._login_user(local_site=False, admin=True)

        local_site = LocalSite.objects.get(pk=1)
        doc = User.objects.get(username='doc')

        applications = self._make_applications(
            users=[self.user, doc],
            local_site=local_site,
            predicate=lambda a: a.local_site is None,
        )

        rsp = self.api_get(get_oauth_app_list_url(),
                           {},
                           expected_mimetype=oauth_app_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(applications,
                         self._applications_from_response(rsp['oauth_apps']))

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_superuser_get_local_site(self):
        """Testing the GET <URL> API with a LocalSite as a superuser"""
        self.user = self._login_user(local_site=False, admin=True)

        local_site = LocalSite.objects.get(pk=1)
        doc = User.objects.get(username='doc')

        applications = self._make_applications(
            users=[self.user, doc],
            local_site=local_site,
            predicate=lambda a: a.local_site == local_site,
        )

        rsp = self.api_get(get_oauth_app_list_url(local_site.name),
                           {},
                           expected_mimetype=oauth_app_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(applications,
                         self._applications_from_response(rsp['oauth_apps']))

    def _applications_from_response(self, item_rsps):
        """Return the Application instances for the given item responses.

        Args:
            item_rsps (list):
                The individual item responses.

        Returns:
            set of reviewboard.oauth.models.Application:
            The matching applications.
        """
        return set(Application.objects.filter(
            pk__in=(item['id'] for item in item_rsps),
        ))

    def _make_applications(self, users, local_site, predicate=None):
        """Create some applications for testing:

        Args:
            users (list of django.contrib.auth.models.User):
                The users to create applications for.

            local_site (reviewboard.site.models.LocalSite):
                A LocalSite.

            predicate (callable, optional):
                An optional callable predicate to filter the results.

        Returns:
            set of reviewboard.oauth.models.Application:
            The created applications.
        """
        applications = set()

        applications.update(
            self.create_oauth_application(u, None, name='%s-app' % u.username)
            for u in users
        )

        applications.update(
            self.create_oauth_application(u, local_site,
                                          name='%s-site-app' % u.username)
            for u in users
        )

        if predicate:
            applications = set(filter(predicate, applications))

        return applications

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        if post_valid_data:
            post_data = {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
            }
        else:
            post_data = {}

        return (get_oauth_app_list_url(local_site_name),
                oauth_app_item_mimetype,
                post_data,
                [])

    def check_post_result(self, user, rsp):
        app = Application.objects.get(pk=rsp['oauth_app']['id'])
        self.compare_item(rsp['oauth_app'], app)

    @webapi_test_template
    def test_post_grant_implicit_no_uris(self):
        """Testing the POST <URL> API with GRANT_IMPLICIT and no URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='',
            grant_type=Application.GRANT_IMPLICIT,
            is_valid=False,
        )

    @webapi_test_template
    def test_post_grant_implicit_uris(self):

        """Testing the POST <URL> API with GRANT_IMPLICIT and URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='https://example.com/',
            grant_type=Application.GRANT_IMPLICIT,
            is_valid=True,
        )

    @webapi_test_template
    def test_post_grant_authorization_code_no_uris(self):
        """Testing the POST <URL> API with GRANT_AUTHORIZATION_CODE and no URIs
        """
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='',
            grant_type=Application.GRANT_AUTHORIZATION_CODE,
            is_valid=False,
        )

    @webapi_test_template
    def test_post_grant_authorization_code_uris(self):
        """Testing the POST <URL> API with GRANT_AUTHORIZATION_CODE and URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='http://example.com',
            grant_type=Application.GRANT_AUTHORIZATION_CODE,
            is_valid=True,
        )

    @webapi_test_template
    def test_post_grant_password_no_uris(self):
        """Testing the POST <URL> API with GRANT_PASSWORD and no URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='',
            grant_type=Application.GRANT_PASSWORD,
            is_valid=True,
        )

    @webapi_test_template
    def test_post_grant_password_uris(self):
        """Testing the POST <URL> API with GRANT_PASSWORD and URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='http://example.com',
            grant_type=Application.GRANT_PASSWORD,
            is_valid=True,
        )

    @webapi_test_template
    def test_post_grant_client_credentials_no_uris(self):
        """Testing the POST <URL> API with GRANT_CLIENT_CREDENTIALS and no URIs
        """
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='',
            grant_type=Application.GRANT_CLIENT_CREDENTIALS,
            is_valid=True,)

    @webapi_test_template
    def test_post_grant_client_credentials_uris(self):
        """Testing the POST <URL> API with GRANT_CLIENT_CREDENTIALS and URIs"""
        self._test_post_redirect_uri_grant_combination(
            redirect_uris='http://example.com',
            grant_type=Application.GRANT_CLIENT_CREDENTIALS,
            is_valid=True,
        )

    @webapi_test_template
    def test_post_set_user(self):
        """Testing the POST <URL> API with user set"""
        rsp = self.api_post(
            get_oauth_app_list_url(),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'user': 'doc',
            },
            expected_status=400,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

        self.assertIn('fields', rsp)
        self.assertIn('user', rsp['fields'])
        self.assertEqual(rsp['fields']['user'],
                         ['You do not have permission to set this field.'])

    @webapi_test_template
    def test_post_set_user_as_superuser(self):
        """Testing the POST <URL> API as a superuser with user set"""
        self._login_user(admin=True)
        rsp = self.api_post(
            get_oauth_app_list_url(),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'user': 'doc',
            },
            expected_mimetype=oauth_app_item_mimetype,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

        app = Application.objects.get(pk=rsp['oauth_app']['id'])
        self.compare_item(rsp['oauth_app'], app)
        self.assertEqual(app.user.username, 'doc')

    @webapi_test_template
    def test_post_set_user_as_superuser_not_exists(self):
        """Testing the POST <URL> API as a superuser with user set as a
        non-existent user
        """
        self._login_user(admin=True)
        rsp = self.api_post(
            get_oauth_app_list_url(),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'user': 'foofoo',
            },
            expected_status=400,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

        self.assertIn('fields', rsp)
        self.assertIn('user', rsp['fields'])
        self.assertEqual(rsp['fields']['user'],
                         ['The user "foofoo" does not exist.'])

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_post_set_user_as_local_site_admin(self):
        """Testing the POST <URL> API as a LocalSite admin with user set"""
        self._login_user(admin=True, local_site=True)
        local_site = LocalSite.objects.get(name=self.local_site_name)
        local_site.users.add(User.objects.get(username='dopey'))

        rsp = self.api_post(
            get_oauth_app_list_url(self.local_site_name),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'user': 'dopey',
            },
            expected_mimetype=oauth_app_item_mimetype,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

        app = Application.objects.get(pk=rsp['oauth_app']['id'])
        self.compare_item(rsp['oauth_app'], app)
        self.assertEqual(app.user.username, 'dopey')

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_post_set_user_as_local_site_admin_with_non_local_site_user(self):
        """Testing the POST <URL> API as a LocalSite admin with user set to a
        non-LocalSite user
        """
        self._login_user(admin=True, local_site=True)

        rsp = self.api_post(
            get_oauth_app_list_url(self.local_site_name),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'user': 'dopey',
            },
            expected_status=400,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

        self.assertIn('fields', rsp)
        self.assertIn('user', rsp['fields'])
        self.assertEqual(
            rsp['fields']['user'],
            ['The user "dopey" does not exist.'],
        )

    @webapi_test_template
    def test_post_set_skip_authorization(self):
        """Testing the POST <URL> API with skip_authorization set"""
        rsp = self.api_post(
            get_oauth_app_list_url(),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'skip_authorization': '1',
            },
            expected_status=400,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

        self.assertIn('fields', rsp)
        self.assertIn('skip_authorization', rsp['fields'])
        self.assertEqual(rsp['fields']['skip_authorization'],
                         ['You do not have permission to set this field.'])

    @webapi_test_template
    def test_post_set_skip_authorization_as_superuser(self):
        """Testing the POST <URL> API as a superuser with skip_authorization"""
        self._login_user(admin=True)

        rsp = self.api_post(
            get_oauth_app_list_url(),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'skip_authorization': '1',
            },
            expected_mimetype=oauth_app_item_mimetype,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

        app = Application.objects.get(pk=rsp['oauth_app']['id'])
        self.compare_item(rsp['oauth_app'], app)
        self.assertEqual(app.skip_authorization, True)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_post_set_skip_authorization_as_local_site_admin(self):
        """Testing the POST <URL> API as a LocalSite admin with
        skip_authorization set
        """
        self._login_user(admin=True, local_site=True)

        rsp = self.api_post(
            get_oauth_app_list_url(self.local_site_name),
            {
                'authorization_grant_type':
                    Application.GRANT_CLIENT_CREDENTIALS,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test-application',
                'redirect_uris': 'https://example.com/oauth/',
                'skip_authorization': '1',
            },
            expected_mimetype=oauth_app_item_mimetype,
        )

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')

        app = Application.objects.get(pk=rsp['oauth_app']['id'])
        self.compare_item(rsp['oauth_app'], app)
        self.assertEqual(app.skip_authorization, True)

    def _test_post_redirect_uri_grant_combination(self, redirect_uris,
                                                  grant_type, is_valid):
        """Test the redirect_uris and grant type are valid or invalid.

        Args:
            redirect_uris (unicode):
                A space-separated list of redirect URIs.

            grant_type (unicode):
                The grant type.

            is_valid (bool):
                Whether or not the given combination is valid. This determines
                the testing done on the response.
        """
        post_data = {
            'authorization_grant_type': grant_type,
            'client_type': Application.CLIENT_PUBLIC,
            'name': 'test-app',
            'redirect_uris': redirect_uris,
            'skip_authorization': '0',
        }

        if is_valid:
            rsp = self.api_post(get_oauth_app_list_url(),
                                post_data,
                                expected_mimetype=oauth_app_item_mimetype)
            self.assertIn('stat', rsp)
            self.assertEqual(rsp['stat'], 'ok')
            self.compare_item(rsp['oauth_app'],
                              Application.objects.get(name='test-app'))
        else:
            rsp = self.api_post(get_oauth_app_list_url(),
                                post_data,
                                expected_status=400)
            self.assertIn('stat', rsp)
            self.assertEqual(rsp['stat'], 'fail')
            self.assertIn('err', rsp)
            self.assertIn('fields', rsp)
            self.assertIn('redirect_uris', rsp['fields'])


@six.add_metaclass(BasicTestsMetaclass)
class ResourceItemTests(ExtraDataItemMixin, BaseWebAPITestCase):
    """Testing the OAuthApplicationResource item APIs."""

    resource = resources.oauth_app
    sample_api_url = 'oauth-apps/<app-id>/'
    fixtures = ['test_users']
    not_owner_status_code = 404
    not_owner_error = DOES_NOT_EXIST

    compare_item = _compare_item

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        app = self.create_oauth_application(user,
                                            with_local_site=with_local_site)

        return (get_oauth_app_item_url(app.pk, local_site_name),
                oauth_app_item_mimetype,
                app)

    @webapi_test_template
    def test_get_without_owner(self):
        """Testing the GET <URL> API without owner"""
        app = self.create_oauth_application(User.objects.get(username='admin'))

        self.api_get(get_oauth_app_item_url(app.pk),
                     expected_status=404)

    @webapi_test_template
    def test_get_without_owner_as_superuser(self):
        """Testing the GET <URL> API without owner as superuser"""
        self.user = self._login_user(admin=True)
        app = self.create_oauth_application(User.objects.get(username='doc'))

        rsp = self.api_get(get_oauth_app_item_url(app.pk),
                           expected_mimetype=oauth_app_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('oauth_app', rsp)
        self.compare_item(rsp['oauth_app'], app)

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_without_local_site(self):
        """Testing the GET <URL> API for an app related to a LocalSite"""
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        app = self.create_oauth_application(
            self.user,
            local_site=LocalSite.objects.get(pk=1))

        rsp = self.api_get(get_oauth_app_item_url(app.pk),
                           expected_status=404)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_with_invalid_local_site(self):
        """Testing the GET <URL> API with an app related to a LocalSite not
        using the LocalSite's API
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        app = self.create_oauth_application(self.user)

        rsp = self.api_get(get_oauth_app_item_url(app.pk, local_site.name),
                           expected_status=404)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')

    @add_fixtures(['test_site'])
    @webapi_test_template
    def test_get_without_owner_as_local_site_admin(self):
        """Testing the GET <URL> API without owner on a LocalSite as a
        LocalSite admin
        """
        local_site = LocalSite.objects.get(pk=1)
        local_site.users.add(self.user)
        app = self.create_oauth_application(self.user, local_site=local_site)
        self.user = self._login_user(admin=True, local_site=True)

        rsp = self.api_get(get_oauth_app_item_url(app.pk, local_site.name),
                           expected_mimetype=oauth_app_item_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('oauth_app', rsp)
        self.compare_item(rsp['oauth_app'], app)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        app = self.create_oauth_application(user,
                                            with_local_site=with_local_site)

        if put_valid_data:
            request_data = {
                'extra_data.fake_key': '',
            }
        else:
            request_data = {
                'user': 'admin',
            }

        return (get_oauth_app_item_url(app.pk, local_site_name),
                oauth_app_item_mimetype,
                request_data,
                app,
                [])

    def check_put_result(self, user, item_rsp, app):
        app = Application.objects.get(pk=app.pk)
        self.compare_item(item_rsp, app)

    @add_fixtures(['test_site'])
    def test_put_re_enable_security_disabled(self):
        """Testing the PUT <URL> API with enabled=1 for an application disabled
        due to security
        """
        self.user = self._login_user(admin=True)
        doc = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(pk=1)
        app = self.create_oauth_application(user=doc, local_site=local_site)

        original_secret = app.client_secret

        local_site.users.remove(doc)

        app = Application.objects.get(pk=app.pk)

        self.assertTrue(app.is_disabled_for_security)
        self.assertEqual(app.user, self.user)
        self.assertEqual(app.original_user, doc)

        rsp = self.api_put(get_oauth_app_item_url(app.pk, local_site.name),
                           {'enabled': '1'},
                           expected_status=400)

        app = Application.objects.get(pk=app.pk)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertIn('fields', rsp)
        self.assertIn('__all__', rsp['fields'])
        self.assertEqual(rsp['fields']['__all__'][0],
                         ApplicationChangeForm.DISABLED_FOR_SECURITY_ERROR)
        self.assertEqual(app.original_user, doc)
        self.assertEqual(app.client_secret, original_secret)

    def test_put_regenerate_secret_key(self):
        """Testing the PUT <URL> API with regenerate_client_secret=1"""
        app = self.create_oauth_application(user=self.user)
        original_secret = app.client_secret

        rsp = self.api_put(get_oauth_app_item_url(app.pk),
                           {'regenerate_client_secret': 1},
                           expected_mimetype=oauth_app_item_mimetype)

        app = Application.objects.get(pk=app.pk)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp['oauth_app'], app)
        self.assertNotEqual(app.client_secret, original_secret)

    @add_fixtures(['test_site'])
    def test_put_regenerate_secret_key_enable(self):
        """Testing the PUT <URL> API with regenerate_secret_key=1 and enabled=1
        """
        self.user = self._login_user(admin=True)
        doc = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(pk=1)
        app = self.create_oauth_application(user=doc, local_site=local_site)

        original_secret = app.client_secret

        local_site.users.remove(doc)

        app = Application.objects.get(pk=app.pk)

        self.assertTrue(app.is_disabled_for_security)
        self.assertEqual(app.user, self.user)
        self.assertEqual(app.original_user, doc)

        rsp = self.api_put(
            get_oauth_app_item_url(app.pk, local_site.name),
            {
                'enabled': '1',
                'regenerate_client_secret': '1',
            },
            expected_mimetype=oauth_app_item_mimetype)

        app = Application.objects.get(pk=app.pk)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['oauth_app']
        self.compare_item(item_rsp, app)
        self.assertNotEqual(item_rsp['client_secret'], original_secret)

        self.assertFalse(app.is_disabled_for_security)
        self.assertIsNone(app.original_user)
        self.assertTrue(app.enabled)
        self.assertNotEqual(app.client_secret, original_secret)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        app = self.create_oauth_application(user=user,
                                            with_local_site=with_local_site)

        return (get_oauth_app_item_url(app.pk, local_site_name),
                [app.pk])

    def check_delete_result(self, user, app_pk):
        self.assertIsNone(get_object_or_none(Application, pk=app_pk))



from __future__ import unicode_literals

import logging
import warnings

from django import template
from django.template import TemplateSyntaxError
from django.template.defaultfilters import escapejs, stringfilter
from django.template.loader import render_to_string
from django.utils import six
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import blocktag
from djblets.util.humanize import humanize_list
from djblets.util.templatetags.djblets_js import json_dumps_items

from reviewboard.accounts.models import Profile, Trophy
from reviewboard.accounts.trophies import UnknownTrophy
from reviewboard.deprecation import RemovedInReviewBoard40Warning
from reviewboard.diffviewer.diffutils import get_displayed_diff_line_ranges
from reviewboard.reviews.actions import get_top_level_actions
from reviewboard.reviews.fields import (get_review_request_field,
                                        get_review_request_fieldset,
                                        get_review_request_fieldsets)
from reviewboard.reviews.markdown_utils import (is_rich_text_default_for_user,
                                                render_markdown,
                                                normalize_text_for_edit)
from reviewboard.reviews.models import (BaseComment, Group,
                                        ReviewRequest, ScreenshotComment,
                                        FileAttachmentComment,
                                        GeneralComment)
from reviewboard.reviews.ui.base import FileAttachmentReviewUI
from reviewboard.site.urlresolvers import local_site_reverse


register = template.Library()


@register.simple_tag(takes_context=True)
def display_review_request_trophies(context, review_request):
    """Returns the HTML for the trophies awarded to a review request."""
    trophy_models = Trophy.objects.get_trophies(review_request)

    if not trophy_models:
        return ''

    trophies = []

    for trophy_model in trophy_models:
        trophy_type_cls = trophy_model.trophy_type

        if trophy_type_cls is not UnknownTrophy:
            try:
                trophy_type = trophy_type_cls()
                text = trophy_type.format_display_text(context['request'],
                                                       trophy_model)

                trophies.append({
                    'image_urls': trophy_type.image_urls,
                    'image_width': trophy_type.image_width,
                    'image_height': trophy_type.image_height,
                    'name': trophy_type.name,
                    'text': text,
                })
            except Exception as e:
                logging.error('Error when rendering trophy %r (%r): %s',
                              trophy_model.pk, trophy_type_cls, e,
                              exc_info=1)

    return render_to_string('reviews/trophy_box.html', {'trophies': trophies})


def _generate_reply_html(context, user, context_id, review, reply, timestamp,
                         last_visited, text, rich_text, anchor_name,
                         use_avatars, extra_context={}):
    """Generate HTML for a single reply.

    Args:
        context (django.template.RequestContext):
            The template context for the page.

        user (django.contrib.auth.models.User):
            The user who is viewing the replies.

        context_id (unicode):
            An internal ID used by the JavaScript code for storing and
            categorizing replies.

        review (reviewboard.reviews.models.review.Review):
            The review being replied to.

        reply (reviewboard.reviews.models.review.Review):
            The reply to the review.

        timestamp (datetime.datetime):
            The timestamp of the reply.

        last_visited (datetime.datetime):
            The last time the user visited the page containing the replies.

        text (unicode):
            The reply text.

        rich_text (bool):
            Whether the reply text is in Markdown format.

        anchor_name (unicode):
            The name of the anchor for the comment, for use in linking to
            this reply.

        use_avatars (bool):
            Whether avatars are enabled on Review Board. This will control
            whether avatars are shown in the replies.

        extra_context (dict):
            Extra template context to include when rendering the page.

    Returns:
        django.utils.safestring.SafeText:
        The HTML for the reply.
    """
    # Note that update() implies push().
    context.update(dict({
        'anchor_name': anchor_name,
        'context_id': context_id,
        'draft': not reply.public,
        'id': reply.pk,
        'reply_is_new': (
            user is not None and
            last_visited is not None and
            reply.is_new_for_user(user, last_visited) and
            not review.is_new_for_user(user, last_visited)),
        'reply_user': reply.user,
        'rich_text': rich_text,
        'text': text,
        'timestamp': timestamp,
        'use_avatars': use_avatars,
    }, **extra_context))

    try:
        return render_to_string('reviews/review_reply.html', context)
    finally:
        context.pop()


@register.simple_tag(takes_context=True)
def comment_replies(context, review, comment, context_id):
    """Render a list of replies to a comment.

    This loads all the replies made to a particular comment and renders
    them in order by timestamp, showing the author of each comment, the
    timestamp, and the text in the appropriate format.

    Args:
        context (django.template.RequestContext):
            The template context for the page.

        review (reviewboard.reviews.models.review.Review):
            The review being replied to.

        comment (reviewboard.reviews.models.base_comment.BaseComment):
            The comment being replied to.

        context_id (unicode):
            An internal ID used by the JavaScript code for storing and
            categorizing replies.

    Returns:
        django.utils.safestring.SafeText:
        The resulting HTML for the replies.
    """
    siteconfig = SiteConfiguration.objects.get_current()
    use_avatars = siteconfig.get('avatars_enabled')
    user = context['request'].user
    last_visited = context.get('last_visited')

    return mark_safe(''.join(
        _generate_reply_html(
            anchor_name='%s%d' % (reply_comment.anchor_prefix,
                                  reply_comment.pk),
            context=context,
            context_id=context_id,
            last_visited=last_visited,
            reply=reply_comment.get_review(),
            review=review,
            rich_text=reply_comment.rich_text,
            text=reply_comment.text,
            timestamp=reply_comment.timestamp,
            use_avatars=use_avatars,
            user=user,
            extra_context={
                'comment_id': reply_comment.pk,
            })
        for reply_comment in comment.public_replies(user)
    ))


@register.simple_tag(takes_context=True)
def review_body_replies(context, review, body_field, context_id):
    """Render a list of replies to a body field of a review.

    This loads all the replies made to a review's header/footer body field and
    renders them in order by timestamp, showing the author of each comment,
    the timestamp, and the text in the appropriate format.

    Args:
        context (django.template.RequestContext):
            The template context for the page.

        review (reviewboard.reviews.models.review.Review):
            The review being replied to.

        body_field (unicode):
            The body field to look up replies to. This can be either
            ``body_top`` or ``body_bottom``.

        context_id (unicode):
            An internal ID used by the JavaScript code for storing and
            categorizing replies.

    Returns:
        django.utils.safestring.SafeText:
        The resulting HTML for the replies.

    Raises:
        django.template.TemplateSyntaxError:
            There was an invalid ``body_field`` provided.
    """
    if body_field not in ('body_top', 'body_bottom'):
        raise TemplateSyntaxError('Invalid body field "%s" provided.'
                                  % body_field)

    siteconfig = SiteConfiguration.objects.get_current()
    use_avatars = siteconfig.get('avatars_enabled')
    user = context['request'].user
    last_visited = context.get('last_visited')
    anchor_field_alias = {
        'body_top': 'header',
        'body_bottom': 'footer',
    }

    replies = getattr(review, 'public_%s_replies' % body_field)(user)

    return mark_safe(''.join(
        _generate_reply_html(
            anchor_name='%s-reply%d' % (anchor_field_alias[body_field],
                                        reply.pk),
            context=context,
            context_id=context_id,
            last_visited=last_visited,
            reply=reply,
            review=review,
            rich_text=getattr(reply, '%s_rich_text' % body_field),
            text=getattr(reply, body_field),
            timestamp=reply.timestamp,
            use_avatars=use_avatars,
            user=user)
        for reply in replies
    ))


@register.inclusion_tag('reviews/review_reply_section.html',
                        takes_context=True)
def reply_section(context, review, comment, context_type, context_id,
                  reply_to_text=''):
    """
    Renders a template for displaying a reply.

    This takes the same parameters as :tag:`reply_list`. The template
    rendered by this function, :template:`reviews/review_reply_section.html`,
    is responsible for invoking :tag:`reply_list` and as such passes these
    variables through. It does not make use of them itself.
    """
    if comment != "":
        if type(comment) is ScreenshotComment:
            context_id += 's'
        elif type(comment) is FileAttachmentComment:
            context_id += 'f'
        elif type(comment) is GeneralComment:
            context_id += 'g'

        context_id += six.text_type(comment.id)

    return {
        'review': review,
        'comment': comment,
        'context_type': context_type,
        'context_id': context_id,
        'user': context.get('user', None),
        'local_site_name': context.get('local_site_name'),
        'reply_to_is_empty': reply_to_text == '',
        'request': context['request'],
        'last_visited': context.get('last_visited'),
    }


@register.simple_tag
def reviewer_list(review_request):
    """
    Returns a humanized list of target reviewers in a review request.
    """
    return humanize_list([group.display_name or group.name
                          for group in review_request.target_groups.all()] +
                         [user.get_full_name() or user.username
                          for user in review_request.target_people.all()])


@register.simple_tag(takes_context=True)
def review_request_actions(context):
    """Render all registered review request actions.

    Args:
        context (django.template.Context):
            The collection of key-value pairs available in the template.

    Returns:
        unicode: The HTML content to be rendered.
    """
    content = []

    for top_level_action in get_top_level_actions():
        try:
            content.append(top_level_action.render(context))
        except Exception:
            logging.exception('Error rendering top-level action %s',
                              top_level_action.action_id)

    return ''.join(content)


@register.simple_tag(takes_context=True)
def child_actions(context):
    """Render all registered child actions.

    Args:
        context (django.template.Context):
            The collection of key-value pairs available in the template.

    Returns:
        unicode: The HTML content to be rendered.
    """
    content = []

    for child_action in context['menu_action']['child_actions']:
        try:
            content.append(child_action.render(context))
        except Exception:
            logging.exception('Error rendering child action %s',
                              child_action.action_id)

    return ''.join(content)


@register.tag
@blocktag(end_prefix='end_')
def for_review_request_field(context, nodelist, review_request_details,
                             fieldset):
    """Loops through all fields in a fieldset.

    This can take a fieldset instance or a fieldset ID.
    """
    s = []

    request = context.get('request')

    if isinstance(fieldset, six.text_type):
        fieldset = get_review_request_fieldset(fieldset)

    for field_cls in fieldset.field_classes:
        try:
            field = field_cls(review_request_details, request=request)
        except Exception as e:
            logging.exception('Error instantiating field %r: %s',
                              field_cls, e)
            continue

        if hasattr(field_cls.should_render, '__call__'):
            warnings.warn('Field %r uses an old style should_render function '
                          'which is deprecated and will be removed in the '
                          'future. This should be converted to a property.'
                          % field_cls,
                          RemovedInReviewBoard40Warning)
            try:
                should_render = field.should_render(field.value)
            except Exception as e:
                logging.exception(
                    'Error running should_render for field %r: %s',
                    field_cls, e)
                should_render = True
        else:
            should_render = field.should_render

        if should_render:
            context.push()
            context['field'] = field
            s.append(nodelist.render(context))
            context.pop()

    return ''.join(s)


@register.tag
@blocktag(end_prefix='end_')
def for_review_request_fieldset(context, nodelist, review_request_details):
    """Loop through all fieldsets.

    Args:
        context (dict):
            The render context.

        nodelist (django.template.NodeList):
            The contents of the template inside the blocktag.

        review_request_details (reviewboard.reviews.models.
                                base_review_request_details.
                                BaseReviewRequestDetails):
            The review request or draft being rendered.

    Returns:
        unicode:
        The rendered tag contents.
    """
    s = []
    is_first = True
    review_request = review_request_details.get_review_request()
    user = context['request'].user

    for fieldset_cls in get_review_request_fieldsets():
        try:
            if not fieldset_cls.is_empty():
                try:
                    fieldset = fieldset_cls(review_request_details)
                except Exception as e:
                    logging.error('Error instantiating ReviewRequestFieldset '
                                  '%r: %s', fieldset_cls, e, exc_info=1)

                # Note that update() implies push().
                context.update({
                    'fieldset': fieldset,
                    'show_fieldset_required': (
                        fieldset.show_required and
                        review_request.status ==
                            ReviewRequest.PENDING_REVIEW and
                        review_request.is_mutable_by(user)),
                    'forloop': {
                        'first': is_first,
                    }
                })

                try:
                    s.append(nodelist.render(context))
                finally:
                    context.pop()

                is_first = False
        except Exception as e:
            logging.error('Error running is_empty for ReviewRequestFieldset '
                          '%r: %s', fieldset_cls, e, exc_info=1)

    return ''.join(s)


@register.tag
@blocktag(end_prefix='end_')
def review_request_field(context, nodelist, review_request_details, field_id):
    """Render a block with a specific review request field.

    Args:
        context (dict):
            The render context.

        nodelist (django.template.NodeList):
            The contents of the template inside the blocktag.

        review_request_details (reviewboard.reviews.models.
                                base_review_request_details.
                                BaseReviewRequestDetails):
            The review request or draft being rendered.

        field_id (unicode):
            The ID of the field to add to the render context.

    Returns:
        unicode:
        The rendered block.
    """
    request = context.get('request')

    try:
        field_cls = get_review_request_field(field_id)
        field = field_cls(review_request_details, request=request)
    except Exception as e:
        logging.exception('Error instantiating field %r: %s',
                          field_id, e)
        return ''

    context.push()

    try:
        context['field'] = field
        return nodelist.render(context)
    finally:
        context.pop()


def _has_usable_review_ui(user, review_request, file_attachment):
    """Return whether there's a usable review UI for a file attachment.

    This will check that a review UI exists for the file attachment and
    that it's enabled for the provided user and review request.

    Args:
        user (django.contrib.auth.models.User):
            The user who would be accessing the review UI.

        review_request (reviewboard.reviews.models.review_request.
                        ReviewRequest):
            The review request that the review UI would be for.

        file_attachment (reviewboard.attachments.models.FileAttachment):
            The file attachment that review UI would review.

    Returns:
        bool:
        ``True`` if a review UI exists and is usable. ``False`` if the
        review UI does not exist, cannot be used, or there's an error when
        checking.
    """
    review_ui = file_attachment.review_ui

    try:
        return (review_ui and
                review_ui.is_enabled_for(user=user,
                                         review_request=review_request,
                                         file_attachment=file_attachment))
    except Exception as e:
        logging.error('Error when calling is_enabled_for '
                      'FileAttachmentReviewUI %r: %s',
                      review_ui, e, exc_info=1)
        return False


@register.assignment_tag
def has_usable_review_ui(user, review_request, file_attachment):
    """Returns whether a review UI is set and can be used."""
    return _has_usable_review_ui(user, review_request, file_attachment)


@register.filter
def bug_url(bug_id, review_request):
    """
    Returns the URL based on a bug number on the specified review request.

    If the repository the review request belongs to doesn't have an
    associated bug tracker, this returns None.
    """
    if (review_request.repository and
        review_request.repository.bug_tracker and
        '%s' in review_request.repository.bug_tracker):
        try:
            return review_request.repository.bug_tracker % bug_id
        except TypeError:
            logging.error("Error creating bug URL. The bug tracker URL '%s' "
                          "is likely invalid." %
                          review_request.repository.bug_tracker)

    return None


@register.simple_tag(takes_context=True)
def star(context, obj):
    """
    Renders the code for displaying a star used for starring items.

    The rendered code should handle click events so that the user can
    toggle the star. The star is rendered by the template
    :template:`reviews/star.html`.

    The passed object must be either a :model:`reviews.ReviewRequest` or
    a :model:`reviews.Group`.
    """
    return render_star(context.get('user', None), obj)


def render_star(user, obj):
    """
    Does the actual work of rendering the star. The star tag is a wrapper
    around this.
    """
    if user.is_anonymous():
        return ""

    profile = None

    if not hasattr(obj, 'starred'):
        try:
            profile = user.get_profile()
        except Profile.DoesNotExist:
            return ""

    if isinstance(obj, ReviewRequest):
        obj_info = {
            'type': 'reviewrequests',
            'id': obj.display_id
        }

        if hasattr(obj, 'starred'):
            starred = obj.starred
        else:
            starred = \
                profile.starred_review_requests.filter(pk=obj.id).count() > 0
    elif isinstance(obj, Group):
        obj_info = {
            'type': 'groups',
            'id': obj.name
        }

        if hasattr(obj, 'starred'):
            starred = obj.starred
        else:
            starred = \
                profile.starred_groups.filter(pk=obj.id).count() > 0
    else:
        raise template.TemplateSyntaxError(
            "star tag received an incompatible object type (%s)" %
            type(obj))

    if starred:
        image_alt = _("Starred")
    else:
        image_alt = _("Click to star")

    return render_to_string('reviews/star.html', {
        'object': obj_info,
        'starred': int(starred),
        'alt': image_alt,
        'user': user,
    })


@register.inclusion_tag('reviews/comment_issue.html',
                        takes_context=True)
def comment_issue(context, review_request, comment, comment_type):
    """
    Renders the code responsible for handling comment issue statuses.
    """

    issue_status = BaseComment.issue_status_to_string(comment.issue_status)
    user = context.get('user', None)

    return {
        'comment': comment,
        'comment_type': comment_type,
        'issue_status': issue_status,
        'review': comment.get_review(),
        'interactive': comment.can_change_issue_status(user),
        'can_verify': comment.can_verify_issue_status(user),
    }


@register.filter
@stringfilter
def pretty_print_issue_status(status):
    """Turns an issue status code into a human-readable status string."""
    return BaseComment.issue_status_to_string(status)


@register.filter
@stringfilter
def issue_status_icon(status):
    """Return an icon name for the issue status.

    Args:
        status (unicode):
            The stored issue status for the comment.

    Returns:
        unicode: The icon name for the issue status.
    """
    if status == BaseComment.OPEN:
        return 'rb-icon-issue-open'
    elif status == BaseComment.RESOLVED:
        return 'rb-icon-issue-resolved'
    elif status == BaseComment.DROPPED:
        return 'rb-icon-issue-dropped'
    elif status in (BaseComment.VERIFYING_RESOLVED,
                    BaseComment.VERIFYING_DROPPED):
        return 'rb-icon-issue-verifying'
    else:
        raise ValueError('Unknown comment issue status "%s"' % status)


@register.filter('render_markdown')
def _render_markdown(text, is_rich_text):
    if is_rich_text:
        return mark_safe(render_markdown(text))
    else:
        return text


@register.simple_tag(takes_context=True)
def expand_fragment_link(context, expanding, tooltip,
                         expand_above, expand_below, text=None):
    """Renders a diff comment fragment expansion link.

    This link will expand the context by the supplied `expanding_above` and
    `expanding_below` values.

    `expanding` is expected to be one of 'above', 'below', or 'line'."""

    lines_of_context = context['lines_of_context']

    image_class = 'rb-icon-diff-expand-%s' % expanding
    expand_pos = (lines_of_context[0] + expand_above,
                  lines_of_context[1] + expand_below)

    return render_to_string('reviews/expand_link.html', {
        'tooltip': tooltip,
        'text': text,
        'comment_id': context['comment'].id,
        'expand_pos': expand_pos,
        'image_class': image_class,
    })


@register.simple_tag(takes_context=True)
def expand_fragment_header_link(context, header):
    """Render a diff comment fragment header expansion link.

    This link expands the context to contain the given line number.
    """
    lines_of_context = context['lines_of_context']
    offset = context['first_line'] - header['line']

    return render_to_string('reviews/expand_link.html', {
        'tooltip': _('Expand to header'),
        'text': format_html('<code>{0}</code>', header['text']),
        'comment_id': context['comment'].id,
        'expand_pos': (lines_of_context[0] + offset,
                       lines_of_context[1]),
        'image_class': 'rb-icon-diff-expand-header',
    })


@register.simple_tag(name='normalize_text_for_edit', takes_context=True)
def _normalize_text_for_edit(context, text, rich_text, escape_js=False):
    text = normalize_text_for_edit(context['request'].user, text, rich_text,
                                   escape_html=not escape_js)

    if escape_js:
        text = escapejs(text)

    return text


@register.simple_tag(takes_context=True)
def rich_text_classname(context, rich_text):
    if rich_text or is_rich_text_default_for_user(context['request'].user):
        return 'rich-text'

    return ''


@register.simple_tag(takes_context=True)
def diff_comment_line_numbers(context, chunks, comment):
    """Render the changed line number ranges for a diff, for use in e-mail.

    This will display the original and patched line ranges covered by a
    comment, transforming the comment's stored virtual line ranges into
    human-readable ranges. It's intended for use in e-mail.

    The template tag's output will list the original line ranges only if
    there are ranges to show, and same with the patched line ranges.

    Args:
        context (django.template.Context):
            The template context.

        chunks (list):
            The list of chunks for the diff.

        comment (reviewboard.reviews.models.diff_comment.Comment):
            The comment containing the line ranges.

    Returns:
        unicode:
        A string representing the line ranges for the comment.
    """
    if comment.first_line is None:
        # Comments without a line number represent the entire file.
        return ''

    orig_range_info, patched_range_info = get_displayed_diff_line_ranges(
        chunks, comment.first_line, comment.last_line)

    if orig_range_info:
        orig_start_linenum, orig_end_linenum = \
            orig_range_info['display_range']

        if orig_start_linenum == orig_end_linenum:
            orig_lines_str = '%s' % orig_start_linenum
            orig_lines_prefix = 'Line'
        else:
            orig_lines_str = '%s-%s' % (orig_start_linenum, orig_end_linenum)
            orig_lines_prefix = 'Lines'
    else:
        orig_lines_str = None
        orig_lines_prefix = None

    if patched_range_info:
        patched_start_linenum, patched_end_linenum = \
            patched_range_info['display_range']

        if patched_start_linenum == patched_end_linenum:
            patched_lines_str = '%s' % patched_start_linenum
            patched_lines_prefix = 'Lines'
        else:
            patched_lines_str = '%s-%s' % (patched_start_linenum,
                                           patched_end_linenum)
            patched_lines_prefix = 'Lines'
    else:
        patched_lines_str = None
        patched_lines_prefix = None

    if orig_lines_str and patched_lines_str:
        return '%s %s (original), %s (patched)' % (
            orig_lines_prefix, orig_lines_str, patched_lines_str)
    elif orig_lines_str:
        return '%s %s (original)' % (orig_lines_prefix, orig_lines_str)
    elif patched_lines_str:
        return '%s %s (patched)' % (patched_lines_prefix, patched_lines_str)
    else:
        return ''


@register.simple_tag(takes_context=True)
def reviewable_page_model_data(context):
    """Output JSON-serialized data for a RB.ReviewablePage model.

    The data will be used by :js:class:`RB.ReviewablePage` in order to
    populate the review request and editor with the necessary state.

    Args:
        context (django.template.RequestContext):
            The current template context.

    Returns:
        unicode:
        The resulting JSON-serialized data. This consists of keys that are
        meant to be injected into an existing dictionary.
    """
    request = context['request']
    user = request.user
    review_request = context['review_request']
    review_request_details = context['review_request_details']
    draft = context['draft']
    close_description = context['close_description']
    close_description_rich_text = context['close_description_rich_text']

    if review_request.local_site:
        local_site_prefix = 's/%s/' % review_request.local_site.name
    else:
        local_site_prefix = ''

    # Build data for the RB.ReviewRequest
    if review_request.status == review_request.PENDING_REVIEW:
        state_data = 'PENDING'
    elif review_request.status == review_request.SUBMITTED:
        state_data = 'CLOSE_SUBMITTED'
    elif review_request.status == review_request.DISCARDED:
        state_data = 'CLOSE_DISCARDED'
    else:
        raise ValueError('Unexpected ReviewRequest.status value "%s"'
                         % review_request.status)

    review_request_data = {
        'id': review_request.display_id,
        'localSitePrefix': local_site_prefix,
        'branch': review_request_details.branch,
        'bugsClosed': review_request_details.get_bug_list(),
        'closeDescription': normalize_text_for_edit(
            user=user,
            text=close_description,
            rich_text=close_description_rich_text,
            escape_html=False),
        'closeDescriptionRichText': close_description_rich_text,
        'description': normalize_text_for_edit(
            user=user,
            text=review_request_details.description,
            rich_text=review_request_details.description_rich_text,
            escape_html=False),
        'descriptionRichText': review_request_details.description_rich_text,
        'hasDraft': draft is not None,
        'lastUpdatedTimestamp': review_request.last_updated,
        'public': review_request.public,
        'reviewURL': review_request.get_absolute_url(),
        'state': state_data,
        'summary': review_request_details.summary,
        'targetGroups': [
            {
                'name': group.name,
                'url': group.get_absolute_url(),
            }
            for group in review_request_details.target_groups.all()
        ],
        'targetPeople': [
            {
                'username': target_user.username,
                'url': local_site_reverse('user',
                                          args=[target_user],
                                          request=request)
            }
            for target_user in review_request_details.target_people.all()
        ],
        'testingDone': normalize_text_for_edit(
            user=user,
            text=review_request_details.testing_done,
            rich_text=review_request_details.testing_done_rich_text,
            escape_html=False),
        'testingDoneRichText': review_request_details.testing_done_rich_text,
    }

    if user.is_authenticated():
        review_request_visit = context['review_request_visit']

        if review_request_visit.visibility == review_request_visit.VISIBLE:
            visibility_data = 'VISIBLE'
        elif review_request_visit.visibility == review_request_visit.ARCHIVED:
            visibility_data = 'ARCHIVED'
        elif review_request_visit.visibility == review_request_visit.MUTED:
            visibility_data = 'MUTED'
        else:
            raise ValueError(
                'Unexpected ReviewRequestVisit.visibility value "%s"'
                % review_request_visit.visibility)

        review_request_data['visibility'] = visibility_data

    repository = review_request.repository

    if repository:
        scmtool = repository.get_scmtool()

        review_request_data['repository'] = {
            'id': repository.pk,
            'name': repository.name,
            'scmtoolName': scmtool.name,
            'requiresBasedir': not scmtool.diffs_use_absolute_paths,
            'requiresChangeNumber': scmtool.supports_pending_changesets,
            'supportsPostCommit': repository.supports_post_commit,
        }

        if repository.bug_tracker:
            review_request_data['bugTrackerURL'] = \
                local_site_reverse(
                    'bug_url',
                    args=[review_request.display_id, '--bug_id--'],
                    request=request)

    if draft:
        review_request_data['submitter'] = {
            'title': draft.submitter.username,
            'url': draft.submitter.get_absolute_url(),
        }

    # Build the data for the RB.ReviewRequestEditor.
    editor_data = {
        'closeDescriptionRenderedText': _render_markdown(
            close_description,
            close_description_rich_text),
        'hasDraft': draft is not None,
        'mutableByUser': context['mutable_by_user'],
        'showSendEmail': context['send_email'],
        'statusMutableByUser': context['status_mutable_by_user'],
    }

    # Build extra data for the RB.ReviewRequest.
    extra_review_request_draft_data = {}

    if draft and draft.changedesc:
        extra_review_request_draft_data.update({
            'changeDescription': normalize_text_for_edit(
                user=user,
                text=draft.changedesc.text,
                rich_text=draft.changedesc.rich_text,
                escape_html=False),
            'changeDescriptionRichText': draft.changedesc.rich_text,
        })

        editor_data['changeDescriptionRenderedText'] = _render_markdown(
            draft.changedesc.text, draft.changedesc.rich_text)

        if draft.diffset:
            extra_review_request_draft_data['interdiffLink'] = \
                local_site_reverse(
                    'view-interdiff',
                    args=[
                        review_request.display_id,
                        draft.diffset.revision - 1,
                        draft.diffset.revision
                    ],
                    request=request)

    # Build the file attachments data for the editor data.
    file_attachments_data = []

    for file_attachment in context.get('file_attachments', []):
        if draft:
            caption = file_attachment.draft_caption
        else:
            caption = file_attachment.caption

        file_attachment_data = {
            'id': file_attachment.pk,
            'loaded': True,
            'caption': caption,
            'downloadURL': file_attachment.get_absolute_url(),
            'filename': file_attachment.filename,
            'revision': file_attachment.attachment_revision,
            'thumbnailHTML': file_attachment.thumbnail,
        }

        if file_attachment.attachment_history_id:
            file_attachment_data['attachmentHistoryID'] = \
                file_attachment.attachment_history_id

        if _has_usable_review_ui(user, review_request, file_attachment):
            file_attachment_data['reviewURL'] = \
                local_site_reverse(
                    'file-attachment',
                    args=[review_request.display_id, file_attachment.pk],
                    request=request)

        file_attachments_data.append(file_attachment_data)

    if file_attachments_data:
        editor_data['fileAttachments'] = file_attachments_data

    # Build the file attachment comments data for the editor data.
    file_attachment_comments_data = {}

    for file_attachment in context.get('all_file_attachments', []):
        review_ui = file_attachment.review_ui

        if not review_ui:
            # For the purposes of serialization, we'll create a dummy ReviewUI.
            review_ui = FileAttachmentReviewUI(file_attachment.review_request,
                                               file_attachment)

        # NOTE: We're setting this here because file attachments serialization
        #       requires this to be set, but we don't necessarily have it set
        #       by this time. We should rethink parts of this down the road,
        #       but it requires dealing with some compatibility issues for
        #       subclasses.
        review_ui.request = request

        file_attachment_comments_data[file_attachment.pk] = \
            review_ui.serialize_comments(file_attachment.get_comments())

    if file_attachment_comments_data:
        editor_data['fileAttachmentComments'] = file_attachment_comments_data

    # And we're done! Assemble it together and chop off the outer dictionary
    # so it can be injected correctly.
    return json_dumps_items({
        'checkForUpdates': True,
        'reviewRequestData': review_request_data,
        'extraReviewRequestDraftData': extra_review_request_draft_data,
        'editorData': editor_data,
    })


@register.simple_tag(takes_context=True)
def render_review_request_entries(context, entries):
    """Render a series of entries on the page.

    Args:
        context (django.template.RequestContext):
            The existing template context on the page.

        entries (list of
                 reviewboard.reviews.detail.BaseReviewRequestPageEntry):
            The entries to render.

    Returns:
        unicode:
        The resulting HTML for the entries.
    """
    request = context['request']

    return ''.join(
        entry.render_to_string(request, context)
        for entry in entries
    )



"""Unit tests for reviewboard.reviews.views.ReviewsDiffViewerView."""

from __future__ import unicode_literals

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class ReviewsDiffViewerViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.ReviewsDiffViewerView."""

    fixtures = ['test_users', 'test_scmtools']

    # Bug 892
    def test_interdiff(self):
        """Testing ReviewsDiffViewerView with interdiffs"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request, revision=1)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/readme',
            dest_file='/readme',
            source_revision='d6613f5',
            dest_detail='5b50866',
            diff=(
                b'diff --git a/readme b/readme\n'
                b'index d6613f5..5b50866 100644\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1,3 @@\n'
                b' Hello there\n'
                b'+\n'
                b'+Oh hi!\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd3\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a new file!\n'
            ))

        diffset = self.create_diffset(review_request, revision=2)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/readme',
            dest_file='/readme',
            source_revision='d6613f5',
            dest_detail='5b50867',
            diff=(
                b'diff --git a/readme b/readme\n'
                b'index d6613f5..5b50867 100644\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1,3 @@\n'
                b' Hello there\n'
                b'+----------\n'
                b'+Oh hi!\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd4\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a diffent version of this new file!\n'
            ))

        response = self.client.get('/r/1/diff/1-2/')

        # Useful for debugging any actual errors here.
        if response.status_code != 200:
            print('Error: %s' % response.context['error'])
            print(response.context['trace'])

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['diff_context']['num_diffs'], 2)

        files = response.context['files']
        self.assertTrue(files)
        self.assertEqual(len(files), 2)

        self.assertEqual(files[0]['depot_filename'], '/newfile')
        self.assertIn('interfilediff', files[0])

        self.assertEqual(files[1]['depot_filename'], '/readme')
        self.assertIn('interfilediff', files[1])

    # Bug 847
    def test_interdiff_new_file(self):
        """Testing ReviewsDiffViewrView with interdiffs containing new files"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request, revision=1)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))

        diffset = self.create_diffset(review_request, revision=2)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd4\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a diffent version of this new file!\n'
            ))

        response = self.client.get('/r/1/diff/1-2/')

        # Useful for debugging any actual errors here.
        if response.status_code != 200:
            print('Error: %s' % response.context['error'])
            print(response.context['trace'])

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['diff_context']['num_diffs'], 2)

        files = response.context['files']
        self.assertTrue(files)
        self.assertEqual(len(files), 1)

        self.assertEqual(files[0]['depot_filename'], '/newfile')
        self.assertIn('interfilediff', files[0])

    def test_with_filenames_option(self):
        """Testing ReviewsDiffViewerView with ?filenames=..."""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff1 = self.create_filediff(diffset,
                                         source_file='src/main/test.c',
                                         dest_file='src/main/test.cpp')
        filediff2 = self.create_filediff(diffset,
                                         source_file='docs/README.txt',
                                         dest_file='docs/README2.txt')
        filediff3 = self.create_filediff(diffset,
                                         source_file='test.txt',
                                         dest_file='test.rst')
        filediff4 = self.create_filediff(diffset,
                                         source_file='/lib/lib.h',
                                         dest_file='/lib/lib.h')
        self.create_filediff(diffset,
                             source_file='unmatched',
                             dest_file='unmatched')

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }),
            {
                'filenames': '*/test.cpp,*.txt,/lib/*',
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['files']
        self.assertEqual({file_info['filediff'] for file_info in files},
                         {filediff1, filediff2, filediff3, filediff4})

    def test_with_filenames_option_normalized(self):
        """Testing ReviewsDiffViewerView with ?filenames=... values normalized
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff1 = self.create_filediff(diffset,
                                         source_file='src/main/test.c',
                                         dest_file='src/main/test.cpp')
        filediff2 = self.create_filediff(diffset,
                                         source_file='docs/README.txt',
                                         dest_file='docs/README2.txt')
        filediff3 = self.create_filediff(diffset,
                                         source_file='test.txt',
                                         dest_file='test.rst')
        filediff4 = self.create_filediff(diffset,
                                         source_file='/lib/lib.h',
                                         dest_file='/lib/lib.h')
        self.create_filediff(diffset,
                             source_file='unmatched',
                             dest_file='unmatched')

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }),
            {
                'filenames': ' ,  , */test.cpp,,,*.txt,/lib/*  ',
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['files']
        self.assertEqual({file_info['filediff'] for file_info in files},
                         {filediff1, filediff2, filediff3, filediff4})



"""Unit tests for reviewboard.accounts.backends.BaseAuthBackend."""

from __future__ import unicode_literals

import re

from reviewboard.accounts.backends import INVALID_USERNAME_CHAR_REGEX
from reviewboard.testing import TestCase


class BaseAuthBackendTests(TestCase):
    """Unit tests for reviewboard.accounts.backends.BaseAuthBackend."""

    def test_invalid_username_char_regex(self):
        """Testing BaseAuthBackend.INVALID_USERNAME_CHAR_REGEX"""
        cases = [
            ('spaces  ', 'spaces'),
            ('spa ces', 'spaces'),
            ('CASES', 'cases'),
            ('CaSeS', 'cases'),
            ('Spec!al', 'specal'),
            ('email@example.com', 'email@example.com'),
            ('da-shes', 'da-shes'),
            ('un_derscores', 'un_derscores'),
            ('mu ^lt&^ipl Es', 'multiples'),
        ]

        for orig, new in cases:
            self.assertEqual(
                re.sub(INVALID_USERNAME_CHAR_REGEX, '', orig).lower(),
                new)



from __future__ import unicode_literals


class UserQueryError(Exception):
    """An error for when a user query fails during user population.

    This error is used by authentication backends implementing the
    :py:meth:`~reviewboard.accounts.backends.base.BaseAuthBackend
    .populate_users` method to report when an error has occurred that should be
    reported back to the caller.
    """

    def __init__(self, msg):
        """Initialize the error.

        Args:
            msg (unicode):
                The error message to display.
        """
        super(Exception, self).__init__(
            'Error while populating users from the auth backend: %s'
            % msg)



"""URLs for integration listing and configuration."""

from __future__ import unicode_literals

from djblets.integrations.urls import build_integration_urlpatterns

from reviewboard.integrations.views import (AdminIntegrationConfigFormView,
                                            AdminIntegrationListView)


#: URL patterns for the administration UI views for configuring integrations.
urlpatterns = build_integration_urlpatterns(
    list_view_cls=AdminIntegrationListView,
    config_form_view_cls=AdminIntegrationConfigFormView)



"""Unit tests for ReviewRequestPageData."""

from __future__ import unicode_literals

from datetime import datetime, timedelta

from django.test.client import RequestFactory
from django.utils import timezone

from reviewboard.reviews.detail import (ChangeEntry,
                                        InitialStatusUpdatesEntry,
                                        ReviewEntry,
                                        ReviewRequestEntry,
                                        ReviewRequestPageData)
from reviewboard.reviews.models import BaseComment, ReviewRequestDraft
from reviewboard.testing import TestCase


class ReviewRequestPageDataTests(TestCase):
    """Unit tests for ReviewRequestPageData."""

    fixtures = ['test_scmtools', 'test_users']

    def test_query_data_pre_etag(self):
        """Testing ReviewRequestPageData.query_data_pre_etag"""
        self._test_query_data_pre_etag_with(
            expected_num_queries=6,
            expect_reviews=True,
            expect_changedescs=True,
            expect_draft=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_only_review_request_entry(self):
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ReviewRequestEntry in entry_classes
        """
        self._test_query_data_pre_etag_with(
            entry_classes=[ReviewRequestEntry],
            expected_num_queries=4,
            expect_reviews=True,
            expect_draft=True)

    def test_query_data_pre_etag_with_only_initial_status_updates_entry(self):
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        InitialStatusUpdatesEntry in entry_classes
        """
        self._test_query_data_pre_etag_with(
            entry_classes=[InitialStatusUpdatesEntry],
            expected_num_queries=4,
            expect_reviews=True,
            expect_status_updates=True)

    def test_query_data_pre_etag_with_only_review_entry(self):
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ReviewEntry in entry_classes
        """
        self._test_query_data_pre_etag_with(
            entry_classes=[ReviewEntry],
            expected_num_queries=3,
            expect_reviews=True)

    def test_query_data_pre_etag_with_only_change_entry(self):
        """Testing ReviewRequestPageData.query_data_pre_etag with only
        ChangeEntry in entry_classes
        """
        self._test_query_data_pre_etag_with(
            entry_classes=[ChangeEntry],
            expected_num_queries=5,
            expect_reviews=True,
            expect_changedescs=True,
            expect_status_updates=True)

    def test_query_data_post_etag(self):
        """Testing ReviewRequestPageData.query_data_post_etag"""
        self._test_query_data_post_etag_with(
            expected_num_queries=18,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_draft=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_review_request_entry(self):
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ReviewRequestEntry in entry_classes
        """
        self._test_query_data_post_etag_with(
            entry_classes=[ReviewRequestEntry],
            expected_num_queries=18,
            expect_draft=True,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_initial_status_updates_entry(self):
        """Testing ReviewRequestPageData.query_data_post_etag with only
        InitialStatusUpdatesEntry in entry_classes
        """
        self._test_query_data_post_etag_with(
            entry_classes=[InitialStatusUpdatesEntry],
            expected_num_queries=18,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_review_entry(self):
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ReviewEntry in entry_classes
        """
        self._test_query_data_post_etag_with(
            entry_classes=[ReviewEntry],
            expected_num_queries=18,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_query_data_post_etag_with_only_change_entry(self):
        """Testing ReviewRequestPageData.query_data_post_etag with only
        ChangeEntry in entry_classes
        """
        self._test_query_data_post_etag_with(
            entry_classes=[ChangeEntry],
            expected_num_queries=18,
            expect_reviews=True,
            expect_file_attachments=True,
            expect_screenshots=True,
            expect_comments=True,
            expect_issues=True)

    def test_get_entries(self):
        """Testing ReviewRequestPageData.get_entries"""
        data = self._build_data()
        data.query_data_pre_etag()
        data.query_data_post_etag()

        entries = data.get_entries()

        self.assertEqual(len(entries['initial']), 1)
        self.assertEqual(len(entries['main']), 4)

        entry = entries['initial'][0]
        self.assertIsInstance(entry, InitialStatusUpdatesEntry)

        entry = entries['main'][0]
        self.assertIsInstance(entry, ReviewEntry)
        self.assertEqual(entry.review, self.review1)

        entry = entries['main'][1]
        self.assertIsInstance(entry, ChangeEntry)
        self.assertEqual(entry.changedesc, self.changedesc1)

        entry = entries['main'][2]
        self.assertIsInstance(entry, ReviewEntry)
        self.assertEqual(entry.review, self.review2)

        entry = entries['main'][3]
        self.assertIsInstance(entry, ChangeEntry)
        self.assertEqual(entry.changedesc, self.changedesc2)

    def _build_data(self, entry_classes=None):
        self._populate_review_request()

        request = RequestFactory().get('/r/1/')
        request.user = self.review_request.submitter

        return ReviewRequestPageData(review_request=self.review_request,
                                     request=request,
                                     entry_classes=entry_classes)

    def _test_query_data_pre_etag_with(self,
                                       entry_classes=None,
                                       expected_num_queries=0,
                                       expect_reviews=False,
                                       expect_changedescs=False,
                                       expect_draft=False,
                                       expect_status_updates=False):
        data = self._build_data(entry_classes=entry_classes)

        with self.assertNumQueries(expected_num_queries):
            data.query_data_pre_etag()

        if expect_reviews:
            self.assertEqual(data.reviews, [self.review2, self.review1])
            self.assertEqual(data.latest_review_timestamp,
                             self.review2.timestamp)
            self.assertEqual(data.diffsets, [self.diffset1, self.diffset2])
            self.assertEqual(
                data.diffsets_by_id,
                {
                    1: self.diffset1,
                    2: self.diffset2,
                })
        else:
            self.assertEqual(data.reviews, [])
            self.assertEqual(data.latest_review_timestamp,
                             datetime.fromtimestamp(0, timezone.utc))
            self.assertEqual(data.diffsets, [])
            self.assertEqual(data.diffsets_by_id, {})

        if expect_changedescs:
            self.assertEqual(data.changedescs, [self.changedesc2,
                                                self.changedesc1])
            self.assertEqual(data.latest_changedesc_timestamp,
                             self.changedesc2.timestamp)
        else:
            self.assertEqual(data.changedescs, [])
            self.assertIsNone(data.latest_changedesc_timestamp)

        if expect_draft:
            self.assertEqual(data.draft, self.draft)
        else:
            self.assertIsNone(data.draft)

        if expect_status_updates:
            self.assertEqual(data.all_status_updates, [self.status_update1,
                                                       self.status_update2])
        else:
            self.assertEqual(data.all_status_updates, [])

    def _test_query_data_post_etag_with(self,
                                        entry_classes=None,
                                        expected_num_queries=0,
                                        expect_draft=False,
                                        expect_reviews=False,
                                        expect_file_attachments=False,
                                        expect_screenshots=False,
                                        expect_comments=False,
                                        expect_issues=False):
        data = self._build_data(entry_classes=entry_classes)

        data.query_data_pre_etag()

        with self.assertNumQueries(expected_num_queries):
            data.query_data_post_etag()

        if expect_reviews:
            self.assertEqual(
                data.reviews_by_id,
                {
                    1: self.review1,
                    2: self.review2,
                })
        else:
            self.assertEqual(data.reviews_by_id, {})

        if expect_draft:
            self.assertEqual(data.review_request_details, self.draft)
        else:
            self.assertEqual(data.review_request_details,
                             self.review_request)

        if expect_file_attachments:
            self.assertEqual(data.active_file_attachments,
                             [self.file_attachment1, self.file_attachment2])
            self.assertEqual(data.all_file_attachments,
                             [self.file_attachment1, self.file_attachment2,
                              self.inactive_file_attachment1])
            self.assertEqual(
                data.file_attachments_by_id,
                {
                    1: self.file_attachment1,
                    2: self.file_attachment2,
                    3: self.inactive_file_attachment1,
                })
        else:
            self.assertEqual(data.active_file_attachments, [])
            self.assertEqual(data.all_file_attachments, [])
            self.assertEqual(data.file_attachments_by_id, {})

        if expect_screenshots:
            self.assertEqual(data.active_screenshots,
                             [self.screenshot1, self.screenshot2])
            self.assertEqual(data.all_screenshots,
                             [self.screenshot1,
                              self.screenshot2,
                              self.inactive_screenshot1])
            self.assertEqual(
                data.screenshots_by_id,
                {
                    1: self.screenshot1,
                    2: self.screenshot2,
                    3: self.inactive_screenshot1,
                })
        else:
            self.assertEqual(data.active_screenshots, [])
            self.assertEqual(data.all_screenshots, [])
            self.assertEqual(data.screenshots_by_id, {})

        if expect_comments:
            self.assertEqual(
                data.all_comments,
                [
                    self.general_comment1,
                    self.general_comment2,
                    self.screenshot_comment1,
                    self.screenshot_comment2,
                    self.file_attachment_comment1,
                    self.file_attachment_comment2,
                    self.diff_comment1,
                    self.diff_comment2,
                ])
        else:
            self.assertEqual(data.all_comments, [])

        if expect_issues:
            self.assertEqual(
                data.issues,
                [
                    self.general_comment1,
                    self.general_comment2,
                    self.file_attachment_comment1,
                    self.file_attachment_comment2,
                    self.diff_comment1,
                    self.diff_comment2,
                ])
            self.assertEqual(
                data.issue_counts,
                {
                    'total': 6,
                    'open': 2,
                    'resolved': 2,
                    'dropped': 2,
                    'verifying': 0,
                })
        else:
            self.assertEqual(data.issues, [])
            self.assertEqual(data.issue_counts, {})

    def _populate_review_request(self):
        now = timezone.now()

        # Create the review request, diffs, attachments, and screenshots.
        self.review_request = self.create_review_request(
            create_repository=True,
            publish=True)

        self.diffset1 = self.create_diffset(self.review_request)
        self.filediff1 = self.create_filediff(self.diffset1)

        self.diffset2 = self.create_diffset(self.review_request)
        self.filediff2 = self.create_filediff(self.diffset2)

        self.file_attachment1 = \
            self.create_file_attachment(self.review_request)
        self.file_attachment2 = \
            self.create_file_attachment(self.review_request)
        self.inactive_file_attachment1 = \
            self.create_file_attachment(self.review_request, active=False)

        self.screenshot1 = self.create_screenshot(self.review_request)
        self.screenshot2 = self.create_screenshot(self.review_request)
        self.inactive_screenshot1 = \
            self.create_screenshot(self.review_request, active=False)

        # Create a draft for this review request.
        self.draft = ReviewRequestDraft.create(self.review_request)

        # Create some reviews.
        self.review1 = self.create_review(self.review_request,
                                          timestamp=now,
                                          publish=True)
        self.general_comment1 = self.create_general_comment(
            self.review1,
            issue_opened=True,
            issue_status=BaseComment.OPEN)
        self.diff_comment1 = self.create_diff_comment(
            self.review1,
            self.filediff1,
            issue_opened=True,
            issue_status=BaseComment.RESOLVED)
        self.file_attachment_comment1 = self.create_file_attachment_comment(
            self.review1,
            self.file_attachment1,
            issue_opened=True,
            issue_status=BaseComment.DROPPED)
        self.screenshot_comment1 = self.create_screenshot_comment(
            self.review1,
            self.screenshot1,
            issue_opened=False)

        self.review2 = self.create_review(self.review_request,
                                          timestamp=now + timedelta(days=3),
                                          publish=True)
        self.general_comment2 = self.create_general_comment(
            self.review2,
            issue_opened=True,
            issue_status=BaseComment.OPEN)
        self.diff_comment2 = self.create_diff_comment(
            self.review2,
            self.filediff2,
            issue_opened=True,
            issue_status=BaseComment.RESOLVED)
        self.file_attachment_comment2 = self.create_file_attachment_comment(
            self.review2,
            self.file_attachment2,
            issue_opened=True,
            issue_status=BaseComment.DROPPED)
        self.screenshot_comment2 = self.create_screenshot_comment(
            self.review2,
            self.screenshot2,
            issue_opened=False)

        # Create some change descriptions.
        self.changedesc1 = self.review_request.changedescs.create(
            timestamp=now + timedelta(days=2),
            public=True)
        self.changedesc2 = self.review_request.changedescs.create(
            timestamp=now + timedelta(days=4),
            public=True)

        # Create some status updates.
        self.status_update1 = self.create_status_update(self.review_request)
        self.status_update2 = self.create_status_update(self.review_request)



from __future__ import unicode_literals

import os
import re
import warnings
from contextlib import contextmanager
from datetime import timedelta

from django.conf import settings
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.files import File
from django.utils import six, timezone
from djblets.testing.testcases import (FixturesCompilerMixin,
                                       TestCase as DjbletsTestCase)
from oauthlib.common import generate_token
from oauth2_provider.models import AccessToken

from reviewboard import scmtools, initialize
from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.admin.siteconfig import load_site_config
from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.differ import DiffCompatVersion
from reviewboard.diffviewer.models import DiffSet, DiffSetHistory, FileDiff
from reviewboard.notifications.models import WebHookTarget
from reviewboard.oauth.models import Application
from reviewboard.reviews.models import (Comment,
                                        FileAttachmentComment,
                                        GeneralComment,
                                        Group,
                                        Review,
                                        ReviewRequest,
                                        ReviewRequestDraft,
                                        Screenshot,
                                        ScreenshotComment,
                                        StatusUpdate)
from reviewboard.scmtools.models import Repository, Tool
from reviewboard.site.models import LocalSite
from reviewboard.webapi.models import WebAPIToken


class TestCase(FixturesCompilerMixin, DjbletsTestCase):
    """The base class for Review Board test cases.

    This class provides a number of convenient functions for creating
    common objects for testing, such as review requests and comments. They're
    populated with default data that can be overridden by the callers.

    This also overcomes an annoyance with default Django unit tests where
    the cache is not cleared across tests, leading to inconsistent results
    and useless testing.
    """
    local_site_name = 'local-site-1'
    local_site_id = 1

    ws_re = re.compile(r'\s+')

    DEFAULT_FILEDIFF_DATA = (
        b'--- README\trevision 123\n'
        b'+++ README\trevision 123\n'
        b'@@ -1 +1 @@\n'
        b'-Hello, world!\n'
        b'+Hello, everybody!\n'
    )

    DEFAULT_GIT_FILEDIFF_DATA = (
        b'diff --git a/README b/README\n'
        b'index 94bdd3e..197009f 100644\n'
        b'--- README\n'
        b'+++ README\n'
        b'@@ -2 +2 @@\n'
        b'-blah blah\n'
        b'+blah!\n'
    )

    def setUp(self):
        super(TestCase, self).setUp()

        initialize()

        self._local_sites = {}

        # Clear the cache so that previous tests don't impact this one.
        cache.clear()

    def shortDescription(self):
        """Returns the description of the current test.

        This changes the default behavior to replace all newlines with spaces,
        allowing a test description to span lines. It should still be kept
        short, though.
        """
        doc = self._testMethodDoc

        if doc is not None:
            doc = doc.split('\n\n', 1)[0]
            doc = self.ws_re.sub(' ', doc).strip()

        return doc

    def get_local_site_or_none(self, name):
        """Returns a LocalSite matching the name, if provided, or None."""
        if name:
            return self.get_local_site(name=name)
        else:
            return None

    def get_local_site(self, name):
        if name not in self._local_sites:
            self._local_sites[name] = LocalSite.objects.get(name=name)

        return self._local_sites[name]

    def create_webapi_token(self, user, note='Sample note',
                            policy={'access': 'rw'},
                            with_local_site=False,
                            **kwargs):
        """Creates a WebAPIToken for testing."""
        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        return WebAPIToken.objects.generate_token(user=user,
                                                  note=note,
                                                  policy=policy,
                                                  local_site=local_site)

    @contextmanager
    def assert_warns(self, cls=DeprecationWarning, message=None):
        """A context manager for asserting code generates a warning.

        This method only supports code which generates a single warning.
        Tests which make use of code generating multiple warnings will
        need to manually catch their warnings.
        """
        with warnings.catch_warnings(record=True) as w:
            # Some warnings such as DeprecationWarning are filtered by
            # default, stop filtering them.
            warnings.simplefilter("always")
            self.assertEqual(len(w), 0)

            yield

            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, cls))

            if message is not None:
                self.assertEqual(message, six.text_type(w[-1].message))

    def create_diff_file_attachment(self, filediff, from_modified=True,
                                    review_request=None,
                                    orig_filename='filename.png',
                                    caption='My Caption',
                                    mimetype='image/png',
                                    **kwargs):
        """Creates a diff-based FileAttachment for testing.

        The FileAttachment is tied to the given FileDiff. It's populated
        with default data that can be overridden by the caller.
        """
        file_attachment = FileAttachment.objects.create_from_filediff(
            filediff=filediff,
            from_modified=from_modified,
            caption=caption,
            orig_filename=orig_filename,
            mimetype=mimetype,
            **kwargs)

        filename = os.path.join(settings.STATIC_ROOT, 'rb', 'images',
                                'logo.png')

        with open(filename, 'r') as f:
            file_attachment.file.save(filename, File(f), save=True)

        if review_request:
            review_request.file_attachments.add(file_attachment)

        return file_attachment

    def create_diffset(self, review_request=None, revision=1, repository=None,
                       draft=False, name='diffset'):
        """Creates a DiffSet for testing.

        The DiffSet defaults to revision 1. This can be overriden by the
        caller.

        DiffSets generally are tied to a ReviewRequest, but it's optional.
        """
        if review_request:
            repository = review_request.repository

        diffset = DiffSet.objects.create(
            name=name,
            revision=revision,
            repository=repository,
            diffcompat=DiffCompatVersion.DEFAULT)

        if review_request:
            if draft:
                review_request_draft = \
                    ReviewRequestDraft.create(review_request)
                review_request_draft.diffset = diffset
                review_request_draft.save()
            else:
                review_request.diffset_history.diffsets.add(diffset)

        return diffset

    def create_diff_comment(self, review, filediff, interfilediff=None,
                            text='My comment', issue_opened=False,
                            issue_status=None, first_line=1, num_lines=5,
                            extra_fields=None, reply_to=None, **kwargs):
        """Create a Comment for testing.

        The comment is tied to the given Review and FileDiff (and, optionally,
        an interfilediff). It's populated with default data that can be
        overridden by the caller.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review associated with the comment.

            filediff (reviewboard.diffviewer.models.FileDiff):
                The FileDiff associated with the comment.

            interfilediff (reviewboard.diffviewer.models.FileDiff, optional):
                The FileDiff used for the end of an interdiff range associated
                with the comment.

            text (unicode):
                The text for the comment.

            issue_opened (bool, optional):
                Whether an issue is to be opened for the comment.

            issue_status (unicode, optional):
                The issue status to set, if an issue is opened. Defaults to
                being an open issue.

            first_line (int, optional):
                The first line (0-based) of the comment range.

            num_lines (int, optional):
                The number of lines in the comment.

            extra_fields (dict, optional):
                Extra data to set on the comment.

            reply_to (reviewboard.reviews.models.diff_comment.Comment,
                      optional):
                The comment this comment replies to.

            **kwargs (dict):
                Additional model attributes to set on the comment.

        Returns:
            reviewboard.reviews.models.diff_comment.Comment:
            The resulting comment.
        """
        if issue_opened and not issue_status:
            issue_status = Comment.OPEN

        comment = Comment(
            filediff=filediff,
            interfilediff=interfilediff,
            first_line=first_line,
            num_lines=num_lines,
            text=text,
            issue_opened=issue_opened,
            issue_status=issue_status,
            reply_to=reply_to,
            **kwargs)

        if extra_fields:
            comment.extra_data = extra_fields

        comment.save()
        review.comments.add(comment)

        return comment

    def create_file_attachment(self, review_request,
                               orig_filename='filename.png',
                               caption='My Caption',
                               draft=False,
                               active=True,
                               **kwargs):
        """Create a FileAttachment for testing.

        The attachment is tied to the given
        :py:class:`~reviewboard.reviews.models.review_request.ReviewRequest`.
        It's populated with default data that can be overridden by the caller.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that ultimately owns the file attachment.

            orig_filename (unicode, optional):
                The filename to use for the file attachment.

            caption (unicode, optional):
                The caption to use for the file attachment.

            draft (bool or
                   reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft):
                A draft to associate the attachment with. This can also be
                a boolean, for legacy reasons, which will attempt to look up
                or create a draft for the review request.

            active (bool):
                Whether this attachment is considered active (not deleted).

            **kwargs (dict):
                Additional fields to set on the attachment.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The resulting file attachment.
        """
        file_attachment = self._create_base_file_attachment(
            caption=caption,
            orig_filename=orig_filename,
            **kwargs)

        if draft:
            if isinstance(draft, ReviewRequestDraft):
                review_request_draft = draft
            else:
                review_request_draft = \
                    ReviewRequestDraft.create(review_request)

            if active:
                attachments = review_request_draft.file_attachments
            else:
                attachments = review_request_draft.inactive_file_attachments
        else:
            if active:
                attachments = review_request.file_attachments
            else:
                attachments = review_request.inactive_file_attachments

        attachments.add(file_attachment)

        return file_attachment

    def create_user_file_attachment(self, user,
                                    caption='My Caption',
                                    with_local_site=False,
                                    local_site_name=None,
                                    local_site=None,
                                    has_file=False,
                                    orig_filename='filename.png',
                                    **kwargs):
        """Create a user FileAttachment for testing.

        The :py:class:`reviewboard.attachments.models.FileAttachment` is tied
        to the given :py:class:`django.contrib.auth.models.User`. It's
        populated with default data that can be overridden by the caller.
        Notably, by default the FileAttachment will be created without a file
        or a local_site.

        Args:
            user (django.contrib.auth.models.User):
                The user who owns the file attachment.

            caption (unicode, optional):
                The caption for the file attachment.

            with_local_site (bool, optional):
                ``True`` if the file attachment should be associated with a
                local site. If this is set, one of ``local_site_name`` or
                ``local_site`` should be provided as well.

            local_site_name (unicode, optional):
                The name of the local site to associate this attachment with.

            local_site (reviewboard.site.models.LocalSite, optional):
                The local site to associate this attachment with.

            has_file (bool, optional):
                ``True`` if an actual file object should be included in the
                model.

            orig_filename (unicode, optional):
                The original name of the file to set in the model.

            kwargs (dict):
                Additional keyword arguments to pass into the FileAttachment
                constructor.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment instance.
        """
        return self._create_base_file_attachment(
            caption=caption,
            user=user,
            has_file=has_file,
            orig_filename=orig_filename,
            with_local_site=with_local_site,
            local_site_name=local_site_name,
            local_site=local_site,
            **kwargs)

    def create_file_attachment_comment(self, review, file_attachment,
                                       diff_against_file_attachment=None,
                                       text='My comment', issue_opened=False,
                                       issue_status=None, extra_fields=None,
                                       reply_to=None, **kwargs):
        """Create a FileAttachmentComment for testing.

        The comment is tied to the given Review and FileAttachment. It's
        populated with default data that can be overridden by the caller.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review associated with the comment.

            file_attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment associated with the comment.

            diff_against_file_attachment (reviewboard.attachments.models.
                                          FileAttachment, optional):
                The file attachment being diff against, for comments on
                attachment diffs.

            text (unicode):
                The text for the comment.

            issue_opened (bool, optional):
                Whether an issue is to be opened for the comment.

            issue_status (unicode, optional):
                The issue status to set, if an issue is opened. Defaults to
                being an open issue.

            extra_fields (dict, optional):
                Extra data to set on the comment.

            reply_to (reviewboard.reviews.models.file_attachment_comment.
                      FileAttachmentComment, optional):
                The comment this comment replies to.

            **kwargs (dict):
                Additional model attributes to set on the comment.

        Returns:
            reviewboard.reviews.models.file_attachment_comment.FileAttachmentComment:
            The resulting comment.
        """
        if issue_opened and not issue_status:
            issue_status = FileAttachmentComment.OPEN

        comment = FileAttachmentComment(
            file_attachment=file_attachment,
            diff_against_file_attachment=diff_against_file_attachment,
            text=text,
            issue_opened=issue_opened,
            issue_status=issue_status,
            reply_to=reply_to,
            **kwargs)

        if extra_fields:
            comment.extra_data = extra_fields

        comment.save()
        review.file_attachment_comments.add(comment)

        return comment

    def create_filediff(self, diffset, source_file='/test-file',
                        dest_file='/test-file', source_revision='123',
                        dest_detail='124', status=FileDiff.MODIFIED,
                        diff=DEFAULT_FILEDIFF_DATA, save=True):
        """Create a FileDiff for testing.

        The FileDiff is tied to the given DiffSet. It's populated with
        default data that can be overridden by the caller.

        Args:
            diffset (reviewboard.diffviewer.models.DiffSet):
                The parent diff set that will own this file.

            source_file (unicode, optional):
                The source filename.

            dest_file (unicode, optional):
                The destination filename, which will be the same as
                ``source_file`` unless the file was moved/renamed/copied.

            source_revision (unicode, optional):
                The source revision.

            dest_detail (unicode, optional):
                The destination revision or other detail as found in the
                parsed diff. This may be a timestamp or some other value.

            status (unicode, optional):
                The status of the file. This is the operation performed
                as indicated in the diff.

            diff (bytes, optional):
                The diff contents.

            save (bool, optional):
                Whether to automatically save the resulting object.

        Returns:
            reviewboard.diffviewer.models.FileDiff:
            The resulting FileDiff.
        """
        filediff = FileDiff(
            diffset=diffset,
            source_file=source_file,
            dest_file=dest_file,
            source_revision=source_revision,
            dest_detail=dest_detail,
            status=status,
            diff=diff)

        if save:
            filediff.save()

        return filediff

    def create_repository(self, with_local_site=False, name='Test Repo',
                          tool_name='Git', path=None, local_site=None,
                          extra_data=None, **kwargs):
        """Create a Repository for testing.

        The Repository may optionally be attached to a
        :py:class:`~reviewboard.site.models.LocalSite`. It's also populated
        with default data that can be overridden by the caller.

        Args:
            with_local_site (bool, optional):
                Whether to create the repository using a Local Site. This
                will choose one based on :py:attr:`local_site_name`.

                If ``local_site`` is provided, this argument is ignored.

            name (unicode, optional):
                The name of the repository.

            tool_name (unicode, optional):
                The name of the registered SCM Tool for the repository.

            path (unicode, optional):
                The path for the repository. If not provided, one will be
                computed.

            local_site (reviewboard.site.models.LocalSite, optional):
                The explicit Local Site to attach.

            extra_data (dict, optional):
                Explicit extra_data to attach to the repository.

            **kwargs (dict):
                Additional fields to set on the repository.

        Returns:
            reviewboard.scmtools.models.Repository:
            The new repository.
        """
        if not local_site:
            if with_local_site:
                local_site = self.get_local_site(name=self.local_site_name)
            else:
                local_site = None

        testdata_dir = os.path.join(os.path.dirname(scmtools.__file__),
                                    'testdata')

        if not path:
            if tool_name in ('Git', 'Test',
                             'TestToolSupportsPendingChangeSets'):
                path = os.path.join(testdata_dir, 'git_repo')
            elif tool_name == 'Subversion':
                path = 'file://' + os.path.join(testdata_dir, 'svn_repo')
            elif tool_name == 'Mercurial':
                path = os.path.join(testdata_dir, 'hg_repo.bundle')
            elif tool_name == 'CVS':
                path = os.path.join(testdata_dir, 'cvs_repo')
            elif tool_name == 'Perforce':
                path = 'localhost:1666'
            else:
                raise NotImplementedError

        repository = Repository(name=name,
                                local_site=local_site,
                                tool=Tool.objects.get(name=tool_name),
                                path=path,
                                **kwargs)

        if extra_data is not None:
            repository.extra_data = extra_data

        repository.save()

        return repository

    def create_review_request(self, with_local_site=False, local_site=None,
                              summary='Test Summary',
                              description='Test Description',
                              testing_done='Testing',
                              submitter='doc',
                              branch='my-branch',
                              local_id=1001,
                              bugs_closed='', status='P', public=False,
                              publish=False, commit_id=None, changenum=None,
                              time_added=None, last_updated=None,
                              repository=None, id=None,
                              create_repository=False,
                              target_people=None,
                              target_groups=None):
        """Create a ReviewRequest for testing.

        The ReviewRequest may optionally be attached to a LocalSite. It's also
        populated with default data that can be overridden by the caller.

        If create_repository is True, a Repository will be created
        automatically. If set, a custom repository cannot be provided.

        The provided submitter may either be a username or a User object.

        If publish is True, ReviewRequest.publish() will be called.
        """
        if not local_site:
            if with_local_site:
                local_site = self.get_local_site(name=self.local_site_name)
            else:
                local_site = None

        if not local_site:
            local_id = None

        if create_repository:
            assert not repository

            repository = \
                self.create_repository(with_local_site=with_local_site)

        if not isinstance(submitter, User):
            submitter = User.objects.get(username=submitter)

        review_request = ReviewRequest(
            summary=summary,
            description=description,
            branch=branch,
            testing_done=testing_done,
            local_site=local_site,
            local_id=local_id,
            submitter=submitter,
            diffset_history=DiffSetHistory.objects.create(),
            repository=repository,
            public=public,
            commit_id=commit_id,
            changenum=changenum,
            bugs_closed=bugs_closed,
            status=status)

        # Set this separately to avoid issues with CounterField updates.
        review_request.id = id

        review_request.save()

        if target_people:
            review_request.target_people = target_people

        if target_groups:
            review_request.target_groups = target_groups

        if publish:
            review_request.publish(review_request.submitter)

        if time_added and last_updated:
            ReviewRequest.objects.filter(pk=review_request.pk).update(
                time_added=time_added,
                last_updated=last_updated)
            review_request.time_added = time_added
            review_request.last_updated = last_updated
        elif time_added:
            ReviewRequest.objects.filter(pk=review_request.pk).update(
                time_added=time_added)
            review_request.time_added = time_added
        elif last_updated:
            ReviewRequest.objects.filter(pk=review_request.pk).update(
                last_updated=last_updated)
            review_request.last_updated = last_updated

        return review_request

    def create_visit(self, review_request, visibility, user='doc',
                     username=None, timestamp=None):
        """Create a ReviewRequestVisit for testing.

        The ReviewRequestVisit is tied to the given ReviewRequest and User.
        It's populated with default data that can be overridden by the caller.

        The provided user may either be a username or a User object.
        """
        if not isinstance(user, six.string_types):
            user = User.objects.get(username=user)

        return ReviewRequestVisit.objects.create(
            review_request=review_request,
            visibility=visibility,
            user=user)

    def create_review(self, review_request, user='dopey',
                      body_top='Test Body Top', body_bottom='Test Body Bottom',
                      ship_it=False, publish=False, timestamp=None, **kwargs):
        """Creates a Review for testing.

        The Review is tied to the given ReviewRequest. It's populated with
        default data that can be overridden by the caller.

        The provided user may either be a username or a User object.

        If publish is True, Review.publish() will be called.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request the review is filed against.

            user (unicode or django.contrib.auth.models.User, optional):
                The username or User object owning the review.

            body_top (unicode, optional):
                The text for the ``body_top`` field.

            body_bottom (unicode, optional):
                The text for the ``body_bottom`` field.

            ship_it (bool, optional):
                The Ship It state for the review.

            publish (bool, optional):
                Whether to publish the review immediately after creation.

            timestamp (datetime.datetime, optional):
                The timestamp for the review.

            **kwargs (dict):
                Additional attributes to set in the review.

        Returns:
            reviewboard.reviews.models.review.Review:
            The resulting review.
        """
        if not isinstance(user, User):
            user = User.objects.get(username=user)

        review = Review.objects.create(
            review_request=review_request,
            user=user,
            body_top=body_top,
            body_bottom=body_bottom,
            ship_it=ship_it,
            **kwargs)

        if publish:
            review.publish()

        if timestamp:
            Review.objects.filter(pk=review.pk).update(timestamp=timestamp)
            review.timestamp = timestamp

        return review

    def create_review_group(self, name='test-group', with_local_site=False,
                            local_site=None, visible=True, invite_only=False,
                            is_default_group=False):
        """Creates a review group for testing.

        The group may optionally be attached to a LocalSite. It's also
        populated with default data that can be overridden by the caller.
        """
        if not local_site and with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)

        return Group.objects.create(
            name=name,
            local_site=local_site,
            visible=visible,
            invite_only=invite_only,
            is_default_group=is_default_group)

    def create_reply(self, review, user='grumpy', body_top='Test Body Top',
                     timestamp=None, publish=False, **kwargs):
        """Create a review reply for testing.

        The reply is tied to the given Review. It's populated with default
        data that can be overridden by the caller.

        To reply to a ``body_top`` or ``body_bottom`` field, pass either
        ``body_top_reply_to=`` or ``body_bottom_reply_to=`` to this method.
        This will be passed to the review's constructor.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review being replied to.

            user (django.contrib.auth.models.User or unicode, optional):
                Either the user model or the username of the user who is
                replying to the review.

            body_top (unicode, optional):
                The body top text.

            timestamp (datetime.datetime, optional):
                The timestamp of the review.

            publish (bool, optional):
                Whether the review should be published. By default it's in
                draft form.

            **kwargs (dict):
                Additional arguments to pass to the
                :py:class:`~reviewboard.reviews.models.review.Review`
                constructor.

        Returns:
            reviewboard.reviews.models.review.Review:
            The resulting review.
        """
        if not isinstance(user, User):
            user = User.objects.get(username=user)

        reply = Review.objects.create(
            review_request=review.review_request,
            user=user,
            body_top=body_top,
            base_reply_to=review,
            **kwargs)

        if publish:
            reply.publish()

        if timestamp:
            Review.objects.filter(pk=reply.pk).update(timestamp=timestamp)
            reply.timestamp = timestamp

        return reply

    def create_screenshot(self, review_request, caption='My caption',
                          draft=False, active=True, **kwargs):
        """Create a Screenshot for testing.

        The screenshot is tied to the given
        :py:class:`~reviewboard.reviews.models.review_request.ReviewRequest`.
        It's populated with default data that can be overridden by the caller.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request that ultimately owns the screenshot.

            caption (unicode, optional):
                The caption to use for the screenshot.

            draft (bool or
                   reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft):
                A draft to associate the screenshot with. This can also be
                a boolean, for legacy reasons, which will attempt to look up
                or create a draft for the review request.

            active (bool):
                Whether this screenshot is considered active (not deleted).

            **kwargs (dict):
                Additional fields to set on the screenshot.

        Returns:
            reviewboard.reviews.models.screenshot.Screenshot:
            The resulting screenshot.
        """
        screenshot = Screenshot(caption=caption, **kwargs)
        filename = os.path.join(settings.STATIC_ROOT, 'rb', 'images',
                                'logo.png')

        with open(filename, 'r') as f:
            screenshot.image.save(filename, File(f), save=True)

        if draft:
            if isinstance(draft, ReviewRequestDraft):
                review_request_draft = draft
            else:
                review_request_draft = \
                    ReviewRequestDraft.create(review_request)

            if active:
                screenshots = review_request_draft.screenshots
            else:
                screenshots = review_request_draft.inactive_screenshots
        else:
            if active:
                screenshots = review_request.screenshots
            else:
                screenshots = review_request.inactive_screenshots

        screenshots.add(screenshot)

        return screenshot

    def create_screenshot_comment(self, review, screenshot, text='My comment',
                                  x=1, y=1, w=5, h=5, issue_opened=False,
                                  issue_status=None, extra_fields=None,
                                  reply_to=None, **kwargs):
        """Create a ScreenshotComment for testing.

        The comment is tied to the given Review and Screenshot. It's
        It's populated with default data that can be overridden by the caller.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review associated with the comment.

            screenshot (reviewboard.reviews.models.screenshot.Screenshot):
                The screenshot associated with the comment.

            text (unicode):
                The text for the comment.

            x (int, optional):
                The X location for the comment on the screenshot.

            y (int, optional):
                The Y location for the comment on the screenshot.

            w (int, optional):
                The width for the comment on the screenshot.

            h (int, optional):
                The height for the comment on the screenshot.

            issue_opened (bool, optional):
                Whether an issue is to be opened for the comment.

            issue_status (unicode, optional):
                The issue status to set, if an issue is opened. Defaults to
                being an open issue.

            extra_fields (dict, optional):
                Extra data to set on the comment.

            reply_to (reviewboard.reviews.models.general_comment.
                      GeneralComment, optional):
                The comment this comment replies to.

            **kwargs (dict):
                Additional model attributes to set on the comment.

        Returns:
            reviewboard.reviews.models.screenshot_comment.ScreenshotComment:
            The resulting comment.
        """
        if issue_opened and not issue_status:
            issue_status = ScreenshotComment.OPEN

        comment = ScreenshotComment(
            screenshot=screenshot,
            text=text,
            x=x,
            y=y,
            w=w,
            h=h,
            issue_opened=issue_opened,
            issue_status=issue_status,
            reply_to=reply_to,
            **kwargs)

        if extra_fields:
            comment.extra_data = extra_fields

        comment.save()
        review.screenshot_comments.add(comment)

        return comment

    def _create_base_file_attachment(self,
                                     caption='My Caption',
                                     orig_filename='filename.png',
                                     has_file=True,
                                     user=None,
                                     with_local_site=False,
                                     local_site_name=None,
                                     local_site=None,
                                     **kwargs):
        """Create a FileAttachment object with the given parameters.

        When creating a
        :py:class:`reviewboard.attachments.models.FileAttachment` that will be
        associated to a review request, a user and local_site should not be
        specified.

        Args:
            caption (unicode, optional):
                The caption for the file attachment.

            orig_filename (unicode, optional):
                The original name of the file to set in the model.

            has_file (bool, optional):
                ``True`` if an actual file object should be included in the
                model.

            user (django.contrib.auth.models.User, optonal):
                The user who owns the file attachment.

            with_local_site (bool, optional):
                ``True`` if the file attachment should be associated with a
                local site. If this is set, one of ``local_site_name`` or
                ``local_site`` should be provided as well.

            local_site_name (unicode, optional):
                The name of the local site to associate this attachment with.

            local_site (reviewboard.site.models.LocalSite, optional):
                The local site to associate this attachment with.

            kwargs (dict):
                Additional keyword arguments to pass into the FileAttachment
                constructor.

        Returns:
            reviewboard.attachments.models.FileAttachment:
            The new file attachment instance.
        """
        if with_local_site:
            local_site = self.get_local_site(name=local_site_name)

        file_attachment = FileAttachment(
            caption=caption,
            user=user,
            uuid='test-uuid',
            local_site=local_site,
            **kwargs)

        if has_file:
            filename = os.path.join(settings.STATIC_ROOT, 'rb', 'images',
                                    'logo.png')

            file_attachment.orig_filename = orig_filename
            file_attachment.mimetype = 'image/png'

            with open(filename, 'r') as f:
                file_attachment.file.save(filename, File(f), save=True)

        file_attachment.save()

        return file_attachment

    def create_general_comment(self, review, text='My comment',
                               issue_opened=False, issue_status=None,
                               extra_fields=None, reply_to=None, **kwargs):
        """Create a GeneralComment for testing.

        The comment is tied to the given Review. It is populated with
        default data that can be overridden by the caller.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review associated with the comment.

            text (unicode):
                The text for the comment.

            issue_opened (bool, optional):
                Whether an issue is to be opened for the comment.

            issue_status (unicode, optional):
                The issue status to set, if an issue is opened. Defaults to
                being an open issue.

            extra_fields (dict, optional):
                Extra data to set on the comment.

            reply_to (reviewboard.reviews.models.general_comment.
                      GeneralComment, optional):
                The comment this comment replies to.

            **kwargs (dict):
                Additional model attributes to set on the comment.

        Returns:
            reviewboard.reviews.models.general_comment.GeneralComment:
            The resulting comment.
        """
        if issue_opened and not issue_status:
            issue_status = GeneralComment.OPEN

        comment = GeneralComment(
            text=text,
            issue_opened=issue_opened,
            issue_status=issue_status,
            reply_to=reply_to,
            **kwargs)

        if extra_fields:
            comment.extra_data = extra_fields

        comment.save()
        review.general_comments.add(comment)

        return comment

    def create_status_update(self, review_request, user='dopey',
                             service_id='service', summary='Status Update',
                             state=StatusUpdate.PENDING,
                             review=None, change_description=None,
                             timestamp=None):
        """Create a status update for testing.

        It is populated with default data that can be overridden by the caller.

        Args:
            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request to associate with the new status update.

            user (django.contrib.auth.models.User or unicode):
                Either the user model or the username of the user who should
                own the status update.

            service_id (unicode):
                The ID to fill in for the new model.

            summary (unicode):
                The summary to fill in for the new model.

            state (unicode):
                The state for the new model. This must be one of the valid
                choices for the state field.

            review (reviewboard.reviews.models.review.Review, optional):
                The review associated with this status update.

            change_description (reviewboard.changedescs.models.
                                ChangeDescription, optional):
                The change description for this status update.

            timestamp (datetime.datetime):
                The timestamp for the status update.

        Returns:
            reviewboard.reviews.models.StatusUpdate:
            The new status update.
        """
        if not isinstance(user, User):
            user = User.objects.get(username=user)

        status_update = StatusUpdate.objects.create(
            review_request=review_request,
            change_description=change_description,
            service_id=service_id,
            summary=summary,
            state=state,
            review=review,
            user=user)

        if timestamp:
            StatusUpdate.objects.filter(pk=status_update.pk).update(
                timestamp=timestamp)
            status_update.timestamp = timestamp

        return status_update

    def create_webhook(self, enabled=False, events=WebHookTarget.ALL_EVENTS,
                       url='http://example.com',
                       encoding=WebHookTarget.ENCODING_JSON,
                       use_custom_content=False, custom_content='',
                       secret='', apply_to=WebHookTarget.APPLY_TO_ALL,
                       repositories=None, with_local_site=False,
                       local_site=None, extra_fields=None):
        """Create a webhook for testing.

        It is populated with default data that can be overridden by the caller.

        Args:
            enabled (bool):
                Whether or not the webhook is enabled when it is created.

            events (unicode):
                A comma-separated list of events that the webhook will trigger
                on.

            url (unicode):
                The URL that requests will be made against.

            encoding (unicode):
                The encoding of the payload to send.

            use_custom_content (bool):
                Determines if custom content will be sent for the payload (if
                ``True``) or if it will be auto-generated (if ``False``).

            custom_content (unicode):
                The custom content to send when ``use_custom_content`` is
                ``True``.

            secret (unicode):
                An HMAC secret to sign the payload with.

            apply_to (unicode):
                The types of repositories the webhook will apply to.

            repositories (list):
                A list of repositories that the webhook will be limited to if
                ``apply_to`` is ``WebHookTarget.APPLY_TO_SELECTED_REPOS``.

            with_local_site (bool):
                Determines if this should be created with a local site.

            local_site (reviewboard.site.models.LocalSite):
                An optional local site. If ``with_local_site`` is ``True`` and
                this argument is ``None``, the local site will be looked up.

            extra_fields (dict):
                Extra data to be imported into the webhook.

        Returns:
            WebHookTarget: A webhook constructed with the given arguments.
        """
        if not local_site:
            if with_local_site:
                local_site = self.get_local_site(name=self.local_site_name)
            else:
                local_site = None

        webhook = WebHookTarget.objects.create(
            enabled=enabled,
            events=events.split(','),
            url=url,
            encoding=encoding,
            use_custom_content=use_custom_content,
            custom_content=custom_content,
            secret=secret,
            apply_to=apply_to,
            local_site=local_site)

        if repositories:
            webhook.repositories = repositories

        if extra_fields:
            webhook.extra_data = extra_fields
            webhook.save(update_fields=['extra_data'])

        return webhook

    def create_oauth_application(
        self, user, local_site=None, with_local_site=False,
        redirect_uris='http://example.com',
        authorization_grant_type=Application.GRANT_CLIENT_CREDENTIALS,
        client_type=Application.CLIENT_PUBLIC,
        **kwargs):
        """Create an OAuth application.

        Args:
            user (django.contrib.auth.models.User):
                The user whom is to own the application.

            local_site (reviewboard.site.models.LocalSite, optional):
                The LocalSite for the application to be associated with, if
                any.

            redirect_uris (unicode, optional):
                A whitespace-separated list of allowable redirect URIs.

            authorization_grant_type (unicode, optional):
                The grant type for the application.

            client_type (unicode, optional):
                The application client type.

            **kwargs (dict):
                Additional keyword arguments to pass to the
                :py:class:`~reviewboard.oauth.models.Application` initializer.

        Returns:
            reviewboard.oauth.models.Application:
            The created application.
        """
        if not local_site:
            if with_local_site:
                local_site = self.get_local_site(self.local_site_name)
            else:
                local_site = None

        return Application.objects.create(
            user=user,
            local_site=local_site,
            authorization_grant_type=authorization_grant_type,
            redirect_uris=redirect_uris,
            client_type=client_type,
            extra_data='{}',
            **kwargs)

    def create_oauth_token(self, application, user, scope='', expires=None,
                           **kwargs):
        """Create an OAuth2 access token for testing.

        Args:
            application (reviewboard.oauth.models.Application):
                The application the token should be associated with.

            user (django.contrib.auth.models.User):
                The user who should own the token.

            scope (unicode, optional):
                The scopes of the token. This argument defaults to the empty
                scope.

            expires (datetime.timedelta, optional):
                How far into the future the token expires. If not provided,
                this argument defaults to one hour.

        Returns:
            oauth2_provider.models.AccessToken:
            The created access token.
        """
        if expires is None:
            expires = timedelta(hours=1)

        return AccessToken.objects.create(
            application=application,
            token=generate_token(),
            expires=timezone.now() + expires,
            scope=scope,
            user=user,
        )

    @contextmanager
    def siteconfig_settings(self, settings, reload_settings=True):
        """Temporarily sets siteconfig settings for a test.

        Args:
            settings (dict):
                The new siteconfig settings to set.

            reload_settings (bool, optional):
                Whether to reload and recompute all settings, applying them
                to Django and other objects.

        Context:
            The current site configuration will contain the new settings for
            this test.
        """
        try:
            with super(TestCase, self).siteconfig_settings(settings):
                if reload_settings:
                    load_site_config()

                yield
        finally:
            if reload_settings:
                load_site_config()



"""Unit tests for reviewboard.accounts.pages.AccountPage."""

from __future__ import unicode_literals

from djblets.registries.errors import RegistrationError

from reviewboard.accounts.forms.pages import AccountPageForm
from reviewboard.accounts.pages import (AccountPage,
                                        get_page_classes,
                                        register_account_page_class,
                                        unregister_account_page_class)
from reviewboard.testing import TestCase


class AccountPageTests(TestCase):
    """Unit tests for reviewboard.accounts.pages.AccountPage."""

    @classmethod
    def setUpClass(cls):
        super(AccountPageTests, cls).setUpClass()

        cls.builtin_pages = set(AccountPage.registry.get_defaults())

    def tearDown(self):
        """Uninitialize this test case."""
        super(AccountPageTests, self).tearDown()
        AccountPage.registry.reset()

    def test_default_pages(self):
        """Testing default list of account pages."""
        self.assertEqual(set(get_page_classes()), self.builtin_pages)

    def test_register_account_page_class(self):
        """Testing register_account_page_class."""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        register_account_page_class(MyPage)

        self.assertEqual(
            set(get_page_classes()),
            self.builtin_pages | {MyPage}
        )

    def test_register_account_page_class_with_duplicate(self):
        """Testing register_account_page_class with duplicate page."""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        register_account_page_class(MyPage)

        with self.assertRaises(RegistrationError):
            register_account_page_class(MyPage)

    def test_unregister_account_page_class(self):
        """Testing unregister_account_page_class."""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        register_account_page_class(MyPage)
        unregister_account_page_class(MyPage)

        self.assertEqual(set(get_page_classes()), self.builtin_pages)

    def test_unregister_unknown_account_page_class(self):
        """Testing unregister_account_page_class with unknown page."""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        with self.assertRaises(AccountPage.registry.lookup_error_class):
            unregister_account_page_class(MyPage)

    def test_add_form_to_page(self):
        """Testing AccountPage.add_form."""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        class MyForm(AccountPageForm):
            form_id = 'test-form'

        register_account_page_class(MyPage)
        MyPage.add_form(MyForm)

        self.assertEqual(MyPage.form_classes, [MyForm])

    def test_add_duplicate_form_to_page(self):
        """Testing AccountPage.add_form with duplicate form ID."""
        class MyForm(AccountPageForm):
            form_id = 'test-form'

        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'
            form_classes = [MyForm]

        register_account_page_class(MyPage)

        with self.assertRaises(RegistrationError):
            MyPage.add_form(MyForm)

        self.assertEqual(MyPage.form_classes, [MyForm])

    def test_remove_form_from_page(self):
        """Testing AccountPage.remove_form."""
        class MyForm(AccountPageForm):
            form_id = 'test-form'

        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'
            form_classes = [MyForm]

        register_account_page_class(MyPage)
        MyPage.remove_form(MyForm)

        self.assertEqual(MyPage.form_classes, [])

    def test_remove_unknown_form_from_page(self):
        """Testing AccountPage.remove_form with unknown form."""
        class MyForm(AccountPageForm):
            form_id = 'test-form'

        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        register_account_page_class(MyPage)

        with self.assertRaises(AccountPage.registry.lookup_error_class):
            MyPage.remove_form(MyForm)

    def test_default_form_classes_for_page(self):
        """Testing AccountPage._default_form_classes persistence"""
        class MyForm(AccountPageForm):
            form_id = 'test-form'

        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'
            form_classes = [MyForm]

        register_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [MyForm])
        unregister_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [])
        register_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [MyForm])

    def test_empty_default_form_classes_for_page(self):
        """Testing AccountPage._default_form_classes with no form_classes"""
        class MyPage(AccountPage):
            page_id = 'test-page'
            page_title = 'Test Page'

        class MyForm(AccountPageForm):
            form_id = 'test-form'

        register_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [])
        MyPage.add_form(MyForm)
        self.assertEqual(MyPage.form_classes, [MyForm])
        unregister_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [])
        register_account_page_class(MyPage)
        self.assertEqual(MyPage.form_classes, [])



"""Standard authentication backend."""

from __future__ import unicode_literals

import logging

from django.conf import settings
from django.contrib.auth import hashers
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import User
from django.utils import six
from django.utils.translation import ugettext_lazy as _
from djblets.db.query import get_object_or_none

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.forms.auth import StandardAuthSettingsForm
from reviewboard.accounts.models import LocalSiteProfile
from reviewboard.site.models import LocalSite


class StandardAuthBackend(BaseAuthBackend, ModelBackend):
    """Authenticate users against the local database.

    This will authenticate a user against their entry in the database, if
    the user has a local password stored. This is the default form of
    authentication in Review Board.

    This backend also handles permission checking for users on LocalSites.
    In Django, this is the responsibility of at least one auth backend in
    the list of configured backends.

    Regardless of the specific type of authentication chosen for the
    installation, StandardAuthBackend will always be provided in the list
    of configured backends. Because of this, it will always be able to
    handle authentication against locally added users and handle
    LocalSite-based permissions for all configurations.
    """

    backend_id = 'builtin'
    name = _('Standard Registration')
    settings_form = StandardAuthSettingsForm
    supports_registration = True
    supports_change_name = True
    supports_change_email = True
    supports_change_password = True

    _VALID_LOCAL_SITE_PERMISSIONS = [
        'hostingsvcs.change_hostingserviceaccount',
        'hostingsvcs.create_hostingserviceaccount',
        'reviews.add_group',
        'reviews.can_change_status',
        'reviews.can_edit_reviewrequest',
        'reviews.can_submit_as_another_user',
        'reviews.can_view_invite_only_groups',
        'reviews.change_default_reviewer',
        'reviews.change_group',
        'reviews.delete_file',
        'reviews.delete_screenshot',
        'scmtools.add_repository',
        'scmtools.change_repository',
    ]

    def authenticate(self, username, password, **kwargs):
        """Authenticate the user.

        This will authenticate the username and return the appropriate User
        object, or None.
        """
        return ModelBackend.authenticate(self, username, password)

    def get_or_create_user(self, username, request):
        """Get an existing user, or create one if it does not exist."""
        return get_object_or_none(User, username=username)

    def update_password(self, user, password):
        """Update the given user's password."""
        user.password = hashers.make_password(password)

    def get_all_permissions(self, user, obj=None):
        """Get a list of all permissions for a user.

        If a LocalSite instance is passed as ``obj``, then the permissions
        returned will be those that the user has on that LocalSite. Otherwise,
        they will be their global permissions.

        It is not legal to pass any other object.
        """
        if obj is not None and not isinstance(obj, LocalSite):
            logging.error('Unexpected object %r passed to '
                          'StandardAuthBackend.get_all_permissions. '
                          'Returning an empty list.',
                          obj)

            if settings.DEBUG:
                raise ValueError('Unexpected object %r' % obj)

            return set()

        if user.is_anonymous():
            return set()

        # First, get the list of all global permissions.
        #
        # Django's ModelBackend doesn't support passing an object, and will
        # return an empty set, so don't pass an object for this attempt.
        permissions = \
            super(StandardAuthBackend, self).get_all_permissions(user)

        if obj is not None:
            # We know now that this is a LocalSite, due to the assertion
            # above.
            if not hasattr(user, '_local_site_perm_cache'):
                user._local_site_perm_cache = {}

            if obj.pk not in user._local_site_perm_cache:
                perm_cache = set()

                try:
                    site_profile = user.get_site_profile(obj)
                    site_perms = site_profile.permissions or {}

                    if site_perms:
                        perm_cache = set([
                            key
                            for key, value in six.iteritems(site_perms)
                            if value
                        ])
                except LocalSiteProfile.DoesNotExist:
                    pass

                user._local_site_perm_cache[obj.pk] = perm_cache

            permissions = permissions.copy()
            permissions.update(user._local_site_perm_cache[obj.pk])

        return permissions

    def has_perm(self, user, perm, obj=None):
        """Get whether or not a user has the given permission.

        If a LocalSite instance is passed as ``obj``, then the permissions
        checked will be those that the user has on that LocalSite. Otherwise,
        they will be their global permissions.

        It is not legal to pass any other object.
        """
        if obj is not None and not isinstance(obj, LocalSite):
            logging.error('Unexpected object %r passed to has_perm. '
                          'Returning False.', obj)

            if settings.DEBUG:
                raise ValueError('Unexpected object %r' % obj)

            return False

        if not user.is_active:
            return False

        if obj is not None:
            if not hasattr(user, '_local_site_admin_for'):
                user._local_site_admin_for = {}

            if obj.pk not in user._local_site_admin_for:
                user._local_site_admin_for[obj.pk] = obj.is_mutable_by(user)

            if user._local_site_admin_for[obj.pk]:
                return perm in self._VALID_LOCAL_SITE_PERMISSIONS

        return super(StandardAuthBackend, self).has_perm(user, perm, obj)
